# Copyright 2015, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio

import numpy as np
import pytest

import spead2
import spead2.send
import spead2.send.asyncio
from spead2.send.asyncio import UdpStream


@pytest.mark.asyncio
class TestUdpStream:
    def setup(self):
        # Make a stream slow enough that we can test async interactions
        config = spead2.send.StreamConfig(rate=5e6)
        self.stream = UdpStream(spead2.ThreadPool(), [('localhost', 8888)], config)
        self.ig = spead2.send.ItemGroup()
        self.ig.add_item(0x1000, 'test', 'Test item', shape=(256 * 1024,), dtype=np.uint8)
        self.ig['test'].value = np.zeros((256 * 1024,), np.uint8)
        self.heap = self.ig.get_heap()

    async def _test_async_flush(self):
        assert self.stream._active > 0
        await self.stream.async_flush()
        assert self.stream._active == 0

    async def test_async_flush(self):
        for i in range(3):
            asyncio.ensure_future(self.stream.async_send_heap(self.heap))
        await self._test_async_flush()

    async def test_async_flush_fail(self):
        """Test async_flush in the case that the last heap sent failed.

        This is arranged by filling up the queue slots first.
        """
        for i in range(5):
            asyncio.ensure_future(self.stream.async_send_heap(self.heap))
        await self._test_async_flush()

    async def test_send_error(self):
        """An error in sending must be reported through the future."""
        # Create a stream with a packet size that is bigger than the likely
        # MTU. It should cause an error.
        stream = UdpStream(
            spead2.ThreadPool(), [("localhost", 8888)],
            spead2.send.StreamConfig(max_packet_size=100000), buffer_size=0)
        with pytest.raises(IOError):
            await stream.async_send_heap(self.heap)


@pytest.mark.asyncio
class TestTcpStream:
    async def test_connect_failed(self):
        thread_pool = spead2.ThreadPool()
        with pytest.raises(IOError):
            await spead2.send.asyncio.TcpStream.connect(thread_pool, [('127.0.0.1', 8887)])



#!/usr/bin/env python

# Copyright 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import copy
import os
import sys

import jinja2
from pycparser import c_ast
from pycparser.c_parser import CParser
from pycparser.c_generator import CGenerator


# The typedefs are arbitrary and just used to allow pycparser to parse the
# code. Note that some functions in infiniband/verbs.h are implemented as
# static inline functions, and so do not get listed here.
IBV_DECLS = '''
typedef unsigned long size_t;
typedef unsigned long uint64_t;

void ibv_ack_cq_events(struct ibv_cq *cq, unsigned int nevents);

struct ibv_pd *ibv_alloc_pd(struct ibv_context *context);

int ibv_close_device(struct ibv_context *context);

struct ibv_comp_channel *ibv_create_comp_channel(struct ibv_context *context);

struct ibv_cq *ibv_create_cq(struct ibv_context *context, int cqe,
                             void *cq_context,
                             struct ibv_comp_channel *channel,
                             int comp_vector);

struct ibv_qp *ibv_create_qp(struct ibv_pd *pd,
                             struct ibv_qp_init_attr *qp_init_attr);

int ibv_dealloc_pd(struct ibv_pd *pd);

int ibv_dereg_mr(struct ibv_mr *mr);

int ibv_destroy_comp_channel(struct ibv_comp_channel *channel);

int ibv_destroy_cq(struct ibv_cq *cq);

int ibv_destroy_qp(struct ibv_qp *qp);

void ibv_free_device_list(struct ibv_device **list);

int ibv_get_cq_event(struct ibv_comp_channel *channel,
                     struct ibv_cq **cq, void **cq_context);

uint64_t ibv_get_device_guid(struct ibv_device *device);

struct ibv_device **ibv_get_device_list(int *num_devices);

struct ibv_context *ibv_open_device(struct ibv_device *device);

int ibv_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
                  int attr_mask);

int ibv_query_device(struct ibv_context *context,
                     struct ibv_device_attr *device_attr);

struct ibv_mr *ibv_reg_mr(struct ibv_pd *pd, void *addr,
                          size_t length, int access);
'''

RDMACM_DECLS = '''
int rdma_bind_addr(struct rdma_cm_id *id, struct sockaddr *addr);

struct rdma_event_channel *rdma_create_event_channel(void);

int rdma_create_id(struct rdma_event_channel *channel,
                   struct rdma_cm_id **id, void *context,
                   enum rdma_port_space ps);

void rdma_destroy_event_channel(struct rdma_event_channel *channel);

int rdma_destroy_id(struct rdma_cm_id *id);
'''

MLX5DV_DECLS = '''
typedef int bool;
typedef unsigned long uint64_t;

bool mlx5dv_is_supported(struct ibv_device *device);

int mlx5dv_query_device(struct ibv_context *ctx_in,
                        struct mlx5dv_context *attrs_out);

struct ibv_wq *mlx5dv_create_wq(struct ibv_context *context,
                                struct ibv_wq_init_attr *wq_init_attr,
                                struct mlx5dv_wq_init_attr *mlx5_wq_attr);

int mlx5dv_init_obj(struct mlx5dv_obj *obj, uint64_t obj_type);
'''


class RenameVisitor(c_ast.NodeVisitor):
    """Renames a function in the AST.

    The function name is stored both in the Decl and in the inner-most TypeDecl
    of the return type. This only handles the latter.
    """

    def __init__(self, old_name, new_name):
        super(RenameVisitor, self).__init__()
        self.old_name = old_name
        self.new_name = new_name

    def visit_TypeDecl(self, node):
        if node.declname == self.old_name:
            node.declname = self.new_name


def rename_func(func, new_name):
    """Return a copy of a function declaration with a new name"""
    func = copy.deepcopy(func)
    RenameVisitor(func.name, new_name).visit(func)
    func.name = new_name
    return func


def make_func_ptr(func, with_name=True):
    """Create a node of pointer-to-function type."""
    if not with_name:
        node = rename_func(func, None)
    else:
        node = copy.deepcopy(func)
    node.type = c_ast.PtrDecl(quals=[], type=node.type)
    return node


def func_args(func):
    """Get list of function argument names"""
    args = [arg.name for arg in func.type.args.params]
    # Handle (void)
    if args == [None]:
        args = []
    return args


class Visitor(c_ast.NodeVisitor):
    """Collects all the function definitions."""

    def __init__(self):
        self.nodes = []

    def visit_Decl(self, node):
        if not isinstance(node.type, c_ast.FuncDecl):
            return
        self.nodes.append(node)


def gen_node(node):
    return CGenerator().visit(node)


class Library:
    """Symbols from a single DSO.

    Parameters
    ----------
    name : str
        Short name used to form symbol names.
    headers : Iterable[str]
        Header files to include.
    soname : str
        Name of library to load.
    guard : str
        Preprocessor expression governing conditional compilation.
    decls : str
        Prototypes of functions to be wrapped.
    wrappers : Iterable[str]
        Functions for which an implementation will be provided in the global
        namespace.
    """

    def __init__(self, name, headers, soname, guard, decls, wrappers=(), fail_log_level='warning'):
        self.name = name
        self.headers = list(headers)
        self.soname = soname
        self.guard = guard
        ast = CParser().parse(decls)
        visitor = Visitor()
        visitor.visit(ast)
        self.nodes = visitor.nodes
        self.wrappers = list(wrappers)
        self.environment = jinja2.Environment(
            autoescape=False,
            trim_blocks=True,
            loader=jinja2.FileSystemLoader(os.path.dirname(__file__))
        )
        self.environment.filters['gen'] = gen_node
        self.environment.filters['rename'] = rename_func
        self.environment.filters['ptr'] = make_func_ptr
        self.environment.filters['args'] = func_args
        self.environment.globals['name'] = self.name
        self.environment.globals['headers'] = self.headers
        self.environment.globals['soname'] = self.soname
        self.environment.globals['guard'] = self.guard
        self.environment.globals['nodes'] = self.nodes
        self.environment.globals['wrappers'] = set(wrappers)
        self.environment.globals['fail_log_level'] = fail_log_level

    def header(self):
        return self.environment.get_template('template.h').render()

    def cxx(self):
        return self.environment.get_template('template.cpp').render()


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('type', choices=['header', 'cxx'])
    parser.add_argument('library', choices=['rdmacm', 'ibv', 'mlx5dv'])
    args = parser.parse_args()

    if args.library == 'rdmacm':
        lib = Library('rdmacm', ['rdma/rdma_cma.h'], 'librdmacm.so.1', 'SPEAD2_USE_IBV',
                      RDMACM_DECLS)
    elif args.library == 'ibv':
        lib = Library('ibv', ['infiniband/verbs.h'], 'libibverbs.so.1', 'SPEAD2_USE_IBV',
                      IBV_DECLS, ['ibv_create_qp', 'ibv_query_device'])
    else:
        lib = Library('mlx5dv', ['infiniband/mlx5dv.h'], 'libmlx5.so.1', 'SPEAD2_USE_MLX5DV',
                      MLX5DV_DECLS, fail_log_level='debug')

    if args.type == 'header':
        print(lib.header())
    else:
        print(lib.cxx())


if __name__ == '__main__':
    main(sys.argv)



# Copyright 2015, 2017-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Receive SPEAD packets and log the contents.

This is both a tool for debugging SPEAD data flows and a demonstrator for the
spead2 package. It thus has many more command-line options than are strictly
necessary, to allow multiple code-paths to be exercised.
"""

import logging
import argparse
import signal
import asyncio

import spead2
import spead2.recv
import spead2.recv.asyncio
from . import cmdline


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('source', nargs='+', help='Sources (filename, host:port or port)')

    group = parser.add_argument_group('Output options')
    group.add_argument('--log', metavar='LEVEL', default='INFO', help='Log level [%(default)s]')
    group.add_argument('--values', action='store_true', help='Show heap values')
    group.add_argument('--descriptors', action='store_true', help='Show descriptors')

    group = parser.add_argument_group('Input options')
    group.add_argument('--max-heaps', type=int, help='Stop receiving after this many heaps')
    group.add_argument('--joint', action='store_true', help='Treat all sources as a single stream')

    group = parser.add_argument_group('Protocol and performance options')
    protocol = cmdline.ProtocolOptions()
    receiver = cmdline.ReceiverOptions(protocol)
    protocol.add_arguments(group)
    receiver.add_arguments(group)

    args = parser.parse_args()
    protocol.notify(parser, args)
    receiver.notify(parser, args)
    return args, receiver


async def run_stream(stream, name, args):
    try:
        item_group = spead2.ItemGroup()
        num_heaps = 0
        while True:
            try:
                if num_heaps == args.max_heaps:
                    break
                heap = await stream.get()
                print(f"Received heap {heap.cnt} on stream {name}")
                num_heaps += 1
                try:
                    if args.descriptors:
                        for raw_descriptor in heap.get_descriptors():
                            descriptor = spead2.Descriptor.from_raw(raw_descriptor, heap.flavour)
                            print('''\
    Descriptor for {0.name} ({0.id:#x})
      description: {0.description}
      format:      {0.format}
      dtype:       {0.dtype}
      shape:       {0.shape}'''.format(descriptor))
                    changed = item_group.update(heap)
                    for (key, item) in changed.items():
                        if args.values:
                            print(key, '=', item.value)
                        else:
                            print(key)
                except ValueError as e:
                    print(f"Error raised processing heap: {e}")
            except (spead2.Stopped, asyncio.CancelledError):
                print(f"Shutting down stream {name} after {num_heaps} heaps")
                stats = stream.stats
                for key in dir(stats):
                    if not key.startswith('_'):
                        print("{}: {}".format(key, getattr(stats, key)))
                break
    finally:
        stream.stop()


def main():
    def make_stream(sources):
        stream = spead2.recv.asyncio.Stream(thread_pool, config, ring_config)
        receiver.add_readers(stream, sources, allow_pcap=True)
        return stream

    def make_coro(sources):
        stream = make_stream(sources)
        return run_stream(stream, sources[0], args), stream

    def stop_streams():
        for stream in streams:
            stream.stop()

    args, receiver = get_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    thread_pool = receiver.make_thread_pool()
    config = receiver.make_stream_config()
    ring_config = receiver.make_ring_stream_config()
    if args.joint:
        coros_and_streams = [make_coro(args.source)]
    else:
        coros_and_streams = [make_coro([source]) for source in args.source]
    coros, streams = zip(*coros_and_streams)
    main_task = asyncio.ensure_future(asyncio.gather(*coros))
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, stop_streams)
    try:
        loop.run_until_complete(main_task)
    except asyncio.CancelledError:
        pass
    loop.close()



# Copyright 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import (
    Text, Union, List, Tuple, Iterator, Iterable, Optional, Sequence, ClassVar, overload
)
import enum
import socket

import spead2
from spead2 import _EndpointList

class Heap:
    def __init__(self, flavour: spead2.Flavour) -> None: ...
    @property
    def flavour(self) -> spead2.Flavour: ...
    @property
    def repeat_pointers(self) -> bool: ...
    @repeat_pointers.setter
    def repeat_pointers(self, value: bool) -> None: ...
    def add_item(self, item: spead2.Item) -> None: ...
    def add_descriptor(self, descriptor: spead2.Descriptor) -> None: ...
    def add_start(self) -> None: ...
    def add_end(self) -> None: ...

class PacketGenerator:
    def __init__(self, heap: Heap, cnt: int, max_packet_size: int) -> None: ...
    def __iter__(self) -> Iterator[bytes]: ...

class RateMethod(enum.Enum):
    SW = ...
    HW = ...
    AUTO = ...

class StreamConfig:
    DEFAULT_MAX_PACKET_SIZE: ClassVar[int]
    DEFAULT_MAX_HEAPS: ClassVar[int]
    DEFAULT_BURST_SIZE: ClassVar[int]
    DEFAULT_BURST_RATE_RATIO: ClassVar[float]
    DEFAULT_RATE_METHOD: ClassVar[RateMethod]

    max_packet_size: int
    rate: float
    burst_size: int
    max_heaps: int
    burst_rate_ratio: float
    rate_method: RateMethod

    def __init__(self, *, max_packet_size: int = ..., rate: float = ...,
                 burst_size: int = ..., max_heaps: int = ...,
                 burst_rate_ratio: float = ...,
                 rate_method: RateMethod = ...) -> None: ...

    @property
    def burst_rate(self) -> float: ...

class _Stream:
    def set_cnt_sequence(self, next: int, step: int) -> None: ...
    @property
    def num_substreams(self) -> int: ...

class _SyncStream(_Stream):
    def send_heap(self, heap: Heap, cnt: int = ..., substream_index = ...) -> None: ...

class _UdpStream:
    DEFAULT_BUFFER_SIZE: ClassVar[int]

    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: str, port: int,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: str = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: str, port: int,
                 config: StreamConfig,
                 buffer_size: int,
                 ttl: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 multicast_group: str, port: int,
                 config: StreamConfig,
                 ttl: int, interface_address: str) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 multicast_group: str, port: int,
                 config: StreamConfig,
                 ttl: int, interface_index: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 socket: socket.socket, hostname: str, port: int,
                 config: StreamConfig = ...) -> None: ...

    # Endpoint list variants
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 endpoints: _EndpointList,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: str = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 endpoints: _EndpointList,
                 config: StreamConfig,
                 buffer_size: int,
                 ttl: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 endpoints: _EndpointList,
                 config: StreamConfig,
                 ttl: int, interface_address: str) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 endpoints: _EndpointList,
                 config: StreamConfig,
                 ttl: int, interface_index: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 socket: socket.socket,
                 endpoints: _EndpointList,
                 config: StreamConfig = ...) -> None: ...

class UdpStream(_UdpStream, _SyncStream):
    pass

class UdpIbvConfig:
    DEFAULT_BUFFER_SIZE: ClassVar[int]
    DEFAULT_MAX_POLL: ClassVar[int]

    endpoints: _EndpointList
    interface_address: str
    buffer_size: int
    ttl: int
    comp_vector: int
    max_poll: int
    memory_regions: list

    def __init__(self, *, endpoints: _EndpointList = ..., interface_address: str = ...,
                 buffer_size: int = ..., ttl: int = ..., comp_vector: int = ...,
                 max_poll: int = ..., memory_regions: list = ...) -> None: ...


class _UdpIbvStream:
    DEFAULT_BUFFER_SIZE: ClassVar[int]
    DEFAULT_MAX_POLL: ClassVar[int]

    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 multicast_group: str, port: int,
                 config: StreamConfig,
                 interface_address: str,
                 buffer_size: int = ..., ttl: int = ...,
                 comp_vector: int = ..., max_pool: int = ...) -> None: ...

    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 config: StreamConfig,
                 udp_ibv_config: UdpIbvConfig) -> None: ...


class UdpIbvStream(_UdpIbvStream, _SyncStream):
    pass

class _TcpStream:
    DEFAULT_BUFFER_SIZE: ClassVar[int]

class TcpStream(_TcpStream, _SyncStream):
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool, socket: socket.socket,
                 config: StreamConfig = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: str, port: int,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: str = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 endpoints: _EndpointList,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: str = ...) -> None: ...

class BytesStream(_SyncStream):
    def getvalue(self) -> bytes: ...
    def __init__(self, thread_pool: spead2.ThreadPool,
                 config: StreamConfig = ...) -> None: ...

class _InprocStream:
    @property
    def queue(self) -> spead2.InprocQueue: ...
    @property
    def queues(self) -> Sequence[spead2.InprocQueue]: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool, queue: spead2.InprocQueue,
                 config: StreamConfig = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool, queues: List[spead2.InprocQueue],
                 config: StreamConfig = ...) -> None: ...

class InprocStream(_InprocStream, _SyncStream):
    pass

class HeapGenerator:
    def __init__(self, item_group: spead2.ItemGroup, descriptor_frequency: Optional[int] = None,
                 flavour: spead2.Flavour = spead2.Flavour()) -> None: ...
    def add_to_heap(self, heap: Heap, descriptors: str = ..., data: str = ...) -> Heap: ...
    def get_heap(self, descriptors: str = ..., data: str = ...) -> Heap: ...
    def get_start(self) -> Heap: ...
    def get_end(self) -> Heap: ...

class ItemGroup(spead2.ItemGroup, HeapGenerator):
    def __init__(self, descriptor_frequency: Optional[int] = None,
                 flavour: spead2.Flavour = ...) -> None: ...



# Copyright 2015, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import binascii
import gc
import math
import struct
import time
import threading
import weakref

import numpy as np
import pytest

import spead2
import spead2.send as send


def hexlify(data):
    """Turns a byte string into human-readable hex dump"""
    if isinstance(data, list):
        return [hexlify(x) for x in data]
    chunks = []
    for i in range(0, len(data), 8):
        part = data[i : min(i + 8, len(data))]
        chunks.append(b':'.join([binascii.hexlify(part[i : i + 1]) for i in range(len(part))]))
    return b' '.join(chunks)


def encode_be(size, value):
    """Encodes `value` as big-endian in `size` bytes"""
    assert size <= 8
    packed = struct.pack('>Q', value)
    return packed[8 - size:]


class Flavour(spead2.Flavour):
    def __init__(self, version, item_pointer_bits, heap_address_bits, bug_compat=0):
        super().__init__(version, item_pointer_bits, heap_address_bits, bug_compat)

    def make_header(self, num_items):
        address_size = self.heap_address_bits // 8
        item_size = 8 - address_size
        return struct.pack(
            '>Q',
            0x5304000000000000 | (address_size << 32) | (item_size << 40) | num_items)

    def make_immediate(self, item_id, value):
        return struct.pack('>Q', 2**63 | (item_id << self.heap_address_bits) | value)

    def make_address(self, item_id, address):
        return struct.pack('>Q', (item_id << self.heap_address_bits) | address)

    def make_shape(self, shape):
        # TODO: extend for bug_compat flavours
        assert not (self.bug_compat &
                    (spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS | spead2.BUG_COMPAT_SHAPE_BIT_1))
        ans = []
        for size in shape:
            if size < 0:
                ans.append(struct.pack('B', 1))
                ans.append(encode_be(self.heap_address_bits // 8, 0))
            else:
                ans.append(struct.pack('B', 0))
                ans.append(encode_be(self.heap_address_bits // 8, size))
        return b''.join(ans)

    def make_format(self, format):
        # TODO: extend for bug_compat flavours
        assert not self.bug_compat & spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS
        ans = []
        for (code, length) in format:
            ans.append(struct.pack('B', ord(code)))
            ans.append(encode_be(8 - self.heap_address_bits // 8, length))
        return b''.join(ans)

    def items_to_bytes(self, items, descriptors=None, max_packet_size=1500, repeat_pointers=False):
        if descriptors is None:
            descriptors = items
        heap = send.Heap(self)
        for descriptor in descriptors:
            heap.add_descriptor(descriptor)
        for item in items:
            heap.add_item(item)
        heap.repeat_pointers = repeat_pointers
        gen = send.PacketGenerator(heap, 0x123456, max_packet_size)
        return list(gen)


def offset_generator(fields):
    offset = 0
    yield offset
    for field in fields:
        offset += len(field)
        yield offset


class TestEncode:
    """Test heap encoding of various data"""

    def setup(self):
        self.flavour = Flavour(4, 64, 48, 0)

    def test_empty(self):
        """An empty heap must still generate a packet"""
        expected = [
            b''.join([
                self.flavour.make_header(5),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 1),
                self.flavour.make_address(spead2.NULL_ID, 0),
                struct.pack('B', 0)])
        ]
        packet = self.flavour.items_to_bytes([])
        assert hexlify(packet) == hexlify(expected)

    def test_lifetime(self):
        """Heap must hold references to item values"""
        item = spead2.Item(id=0x2345, name='name', description='description',
                           shape=(2, 3), dtype=np.uint16)
        item.value = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        weak = weakref.ref(item.value)
        heap = send.Heap(self.flavour)
        heap.add_item(item)
        del item
        packets = list(send.PacketGenerator(heap, 0x123456, 1472))   # noqa: F841
        assert weak() is not None
        del heap
        # pypy needs multiple gc passes to wind it all up
        for i in range(10):
            gc.collect()
        assert weak() is None

    def make_descriptor_numpy(self, id, name, description, shape, dtype_str, fortran_order):
        payload_fields = [
            b'name',
            b'description',
            b'',
            self.flavour.make_shape(shape),
            "{{'descr': {!r}, 'fortran_order': {!r}, 'shape': {!r}}}".format(
                str(dtype_str), bool(fortran_order), tuple(shape)).encode()
        ]
        payload = b''.join(payload_fields)
        offsets = offset_generator(payload_fields)
        descriptor = b''.join([
            self.flavour.make_header(10),
            self.flavour.make_immediate(spead2.HEAP_CNT_ID, 1),
            self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
            self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
            self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
            self.flavour.make_immediate(spead2.DESCRIPTOR_ID_ID, id),
            self.flavour.make_address(spead2.DESCRIPTOR_NAME_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_DESCRIPTION_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_FORMAT_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_SHAPE_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_DTYPE_ID, next(offsets)),
            payload
        ])
        return descriptor

    def make_descriptor_fallback(self, id, name, description, shape, format):
        payload_fields = [
            b'name',
            b'description',
            self.flavour.make_format(format),
            self.flavour.make_shape(shape),
        ]
        payload = b''.join(payload_fields)
        offsets = offset_generator(payload_fields)
        descriptor = b''.join([
            self.flavour.make_header(9),
            self.flavour.make_immediate(spead2.HEAP_CNT_ID, 1),
            self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
            self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
            self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
            self.flavour.make_immediate(spead2.DESCRIPTOR_ID_ID, id),
            self.flavour.make_address(spead2.DESCRIPTOR_NAME_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_DESCRIPTION_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_FORMAT_ID, next(offsets)),
            self.flavour.make_address(spead2.DESCRIPTOR_SHAPE_ID, next(offsets)),
            payload
        ])
        return descriptor

    def test_numpy_simple(self):
        """A single numpy-format item with descriptor"""
        id = 0x2345
        shape = (2, 3)
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        payload_fields = [
            self.make_descriptor_numpy(id, 'name', 'description', shape, '<u2', False),
            struct.pack('<6H', 6, 7, 8, 10, 11, 12000)
        ]
        payload = b''.join(payload_fields)
        offsets = offset_generator(payload_fields)
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
                self.flavour.make_address(spead2.DESCRIPTOR_ID, next(offsets)),
                self.flavour.make_address(id, next(offsets)),
                payload
            ])
        ]

        item = spead2.Item(id=id, name='name', description='description',
                           shape=shape, dtype=np.uint16)
        item.value = data
        packet = self.flavour.items_to_bytes([item])
        assert hexlify(packet) == hexlify(expected)

    def test_numpy_noncontiguous(self):
        """A numpy item with a discontiguous item value is sent correctly"""
        id = 0x2345
        shape = (2, 3)
        store = np.array([[6, 7, 8, 0, 1], [10, 11, 12000, 2, 3], [9, 9, 9, 9, 9]], dtype=np.uint16)
        data = store[:2, :3]
        payload = struct.pack('<6H', 6, 7, 8, 10, 11, 12000)
        expected = [
            b''.join([
                self.flavour.make_header(5),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
                self.flavour.make_address(id, 0),
                payload
            ])
        ]
        item = spead2.Item(id=id, name='name', description='description',
                           shape=shape, dtype=np.uint16)
        item.value = data
        packet = self.flavour.items_to_bytes([item], [])
        assert hexlify(packet) == hexlify(expected)

    def test_numpy_fortran_order(self):
        """A numpy item with Fortran-order descriptor must be sent in Fortran order"""
        id = 0x2345
        shape = (2, 3)
        data = np.array([[6, 7, 8], [10, 11, 12000]], order='F')
        assert data.flags.c_contiguous is False
        payload_fields = [
            self.make_descriptor_numpy(id, 'name', 'description', shape, '<u2', True),
            struct.pack('<6H', 6, 10, 7, 11, 8, 12000)
        ]
        payload = b''.join(payload_fields)
        offsets = offset_generator(payload_fields)
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
                self.flavour.make_address(spead2.DESCRIPTOR_ID, next(offsets)),
                self.flavour.make_address(id, next(offsets)),
                payload
            ])
        ]
        item = spead2.Item(id=id, name='name', description='description',
                           shape=shape, dtype=np.uint16, order='F')
        item.value = data
        packet = self.flavour.items_to_bytes([item])
        assert hexlify(packet) == hexlify(expected)

    def test_fallback_types(self):
        """Send an array with mixed types and strange packing"""
        id = 0x2345
        format = [('b', 1), ('i', 7), ('c', 8), ('f', 32)]
        shape = (2,)
        data = [(True, 17, 'y', 1.0), (False, -23.0, 'n', -1.0)]
        payload_fields = [
            self.make_descriptor_fallback(id, 'name', 'description', shape, format),
            b'\x91y\x3F\x80\x00\x00' + b'\x69n\xBF\x80\x00\x00'
        ]
        payload = b''.join(payload_fields)
        offsets = offset_generator(payload_fields)
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
                self.flavour.make_address(spead2.DESCRIPTOR_ID, next(offsets)),
                self.flavour.make_address(id, next(offsets)),
                payload
            ])
        ]
        item = spead2.Item(id=id, name='name', description='description',
                           shape=shape, format=format)
        item.value = data
        packet = self.flavour.items_to_bytes([item])
        assert hexlify(packet) == hexlify(expected)

    def test_small_fixed(self):
        """Sending a small item with fixed shape must use an immediate."""
        id = 0x2345
        data = 0x7654
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 1),
                self.flavour.make_immediate(id, data),
                self.flavour.make_address(spead2.NULL_ID, 0),
                struct.pack('B', 0)
            ])
        ]
        item = spead2.Item(id=id, name='name', description='description',
                           shape=(), format=[('u', 16)])
        item.value = data
        packet = self.flavour.items_to_bytes([item], [])
        assert hexlify(packet) == hexlify(expected)

    def test_small_variable(self):
        """Sending a small item with dynamic shape must not use an immediate."""
        id = 0x2345
        shape = (1, None)
        data = np.array([[4, 5]], dtype=np.uint8)
        payload = struct.pack('>2B', 4, 5)
        expected = [
            b''.join([
                self.flavour.make_header(5),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, len(payload)),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, len(payload)),
                self.flavour.make_address(id, 0),
                payload
            ])
        ]
        item = spead2.Item(id=id, name='name', description='description',
                           shape=shape, format=[('u', 8)])
        item.value = data
        packet = self.flavour.items_to_bytes([item], [])
        assert hexlify(packet) == hexlify(expected)

    def test_numpy_zero_length(self):
        """A zero-length numpy type raises :exc:`ValueError`"""
        with pytest.raises(ValueError):
            spead2.Item(id=0x2345, name='name', description='description',
                        shape=(), dtype=np.str_)

    def test_fallback_zero_length(self):
        """A zero-length type raises :exc:`ValueError`"""
        with pytest.raises(ValueError):
            spead2.Item(id=0x2345, name='name', description='description',
                        shape=(), format=[('u', 0)])

    def test_start(self):
        """Tests sending a start-of-stream marker."""
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_START),
                self.flavour.make_address(spead2.NULL_ID, 0),
                struct.pack('B', 0)
            ])
        ]
        heap = send.Heap(self.flavour)
        heap.add_start()
        packet = list(send.PacketGenerator(heap, 0x123456, 1500))
        assert hexlify(packet) == hexlify(expected)

    def test_end(self):
        """Tests sending an end-of-stream marker."""
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_STOP),
                self.flavour.make_address(spead2.NULL_ID, 0),
                struct.pack('B', 0)
            ])
        ]
        heap = send.Heap(self.flavour)
        heap.add_end()
        packet = list(send.PacketGenerator(heap, 0x123456, 1500))
        assert hexlify(packet) == hexlify(expected)

    def test_replicate_pointers(self):
        """Tests sending a heap with replicate_pointers set to true"""
        id = 0x2345
        data = np.arange(32, dtype=np.uint8)
        item1 = spead2.Item(id=id, name='item1', description='addressed item',
                            shape=data.shape, dtype=data.dtype, value=data)
        item2 = spead2.Item(id=id + 1, name='item2', description='inline item',
                            shape=(), format=[('u', self.flavour.heap_address_bits)],
                            value=0xdeadbeef)
        expected = [
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 32),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 16),
                self.flavour.make_address(id, 0),
                self.flavour.make_immediate(id + 1, 0xdeadbeef),
                data.tobytes()[0:16]
            ]),
            b''.join([
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, 0x123456),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 32),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 16),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 16),
                self.flavour.make_address(id, 0),
                self.flavour.make_immediate(id + 1, 0xdeadbeef),
                data.tobytes()[16:32]
            ])
        ]
        packets = self.flavour.items_to_bytes([item1, item2], [], max_packet_size=72,
                                              repeat_pointers=True)
        assert hexlify(packets) == hexlify(expected)


class TestStreamConfig:
    def test_default_construct(self):
        config = send.StreamConfig()
        assert config.max_packet_size == config.DEFAULT_MAX_PACKET_SIZE
        assert config.rate == 0.0
        assert config.burst_size == config.DEFAULT_BURST_SIZE
        assert config.max_heaps == config.DEFAULT_MAX_HEAPS
        assert config.burst_rate_ratio == config.DEFAULT_BURST_RATE_RATIO
        assert config.rate_method == config.DEFAULT_RATE_METHOD

    def test_setters(self):
        config = send.StreamConfig()
        config.max_packet_size = 1234
        config.rate = 1e9
        config.burst_size = 12345
        config.max_heaps = 5
        config.burst_rate_ratio = 1.5
        config.rate_method = send.RateMethod.SW
        assert config.max_packet_size == 1234
        assert config.rate == 1e9
        assert config.burst_size == 12345
        assert config.max_heaps == 5
        assert config.burst_rate_ratio == 1.5
        assert config.rate_method == send.RateMethod.SW
        assert config.burst_rate == 1.5e9

    def test_construct_kwargs(self):
        config = send.StreamConfig(max_packet_size=1234, max_heaps=5)
        assert config.max_packet_size == 1234
        assert config.max_heaps == 5
        assert config.rate == 0.0

    def test_bad_rate(self):
        with pytest.raises(ValueError):
            send.StreamConfig(rate=-1.0)
        with pytest.raises(ValueError):
            send.StreamConfig(rate=math.nan)

    def test_bad_max_heaps(self):
        with pytest.raises(TypeError):
            send.StreamConfig(max_heaps=-1)
        with pytest.raises(ValueError):
            send.StreamConfig(max_heaps=0)


class TestStream:
    def setup(self):
        # A slow stream, so that we can test overflowing the queue
        self.flavour = Flavour(4, 64, 48, 0)
        self.stream = send.BytesStream(
            spead2.ThreadPool(),
            send.StreamConfig(rate=1e6, max_heaps=2))
        # A large heap
        ig = send.ItemGroup(flavour=self.flavour)
        ig.add_item(0x1000, 'test', 'A large item', shape=(256 * 1024,), dtype=np.uint8)
        ig['test'].value = np.zeros((256 * 1024,), np.uint8)
        self.heap = ig.get_heap()
        self.threads = []

    def teardown(self):
        for thread in self.threads:
            thread.join()

    def test_overflow(self):
        # Use threads to fill up the first two slots. This is necessary because
        # send_heap is synchronous.
        for i in range(2):
            thread = threading.Thread(target=lambda: self.stream.send_heap(self.heap))
            thread.start()
            self.threads.append(thread)
        # There shouldn't be room now. The sleep is an ugly hack to wait for
        # the threads to enqueue their heaps.
        time.sleep(0.05)
        with pytest.raises(IOError):
            self.stream.send_heap(self.heap)

    def test_send_error(self):
        """An error in sending must be reported."""
        # Create a stream with a packet size that is bigger than the likely
        # MTU. It should cause an error.
        stream = send.UdpStream(
            spead2.ThreadPool(), [("localhost", 8888)],
            send.StreamConfig(max_packet_size=100000), buffer_size=0)
        with pytest.raises(IOError):
            stream.send_heap(self.heap)

    def test_send_explicit_cnt(self):
        """An explicit set heap ID must be respected, and not increment the
        implicit sequence.

        The implicit sequencing is also tested, including wrapping
        """
        ig = send.ItemGroup(flavour=self.flavour)
        self.stream.send_heap(ig.get_start())
        self.stream.set_cnt_sequence(0x1111111111, 0x1234512345)
        self.stream.send_heap(ig.get_start())
        self.stream.send_heap(ig.get_start(), 0x9876543210ab)
        self.stream.send_heap(ig.get_start())
        self.stream.set_cnt_sequence(2**48 - 1, 1)
        self.stream.send_heap(ig.get_start())
        self.stream.send_heap(ig.get_start())
        expected_cnts = [1, 0x1111111111, 0x9876543210ab, 0x2345623456,
                         2**48 - 1, 0]
        expected = b''
        for cnt in expected_cnts:
            expected = b''.join([
                expected,
                self.flavour.make_header(6),
                self.flavour.make_immediate(spead2.HEAP_CNT_ID, cnt),
                self.flavour.make_immediate(spead2.HEAP_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.PAYLOAD_OFFSET_ID, 0),
                self.flavour.make_immediate(spead2.PAYLOAD_LENGTH_ID, 1),
                self.flavour.make_immediate(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_START),
                self.flavour.make_address(spead2.NULL_ID, 0),
                struct.pack('B', 0)
            ])
        assert hexlify(self.stream.getvalue()) == hexlify(expected)

    def test_invalid_cnt(self):
        """An explicit heap ID that overflows must raise an error."""
        ig = send.ItemGroup(flavour=self.flavour)
        with pytest.raises(IOError):
            self.stream.send_heap(ig.get_start(), 2**48)


class TestTcpStream:
    def test_failed_connect(self):
        with pytest.raises(IOError):
            send.TcpStream(spead2.ThreadPool(), [('127.0.0.1', 8887)])


class TestInprocStream:
    def setup(self):
        self.flavour = Flavour(4, 64, 48, 0)
        self.queue = spead2.InprocQueue()
        self.stream = send.InprocStream(spead2.ThreadPool(), [self.queue])

    def test_stopped_queue(self):
        self.queue.stop()
        ig = send.ItemGroup()
        with pytest.raises(IOError):
            self.stream.send_heap(ig.get_end())


@pytest.mark.skipif(not hasattr(spead2, 'IbvContext'), reason='IBV support not compiled in')
class TestUdpIbvConfig:
    def test_default_construct(self):
        config = send.UdpIbvConfig()
        assert config.endpoints == []
        assert config.interface_address == ''
        assert config.buffer_size == send.UdpIbvConfig.DEFAULT_BUFFER_SIZE
        assert config.ttl == 1
        assert config.comp_vector == 0
        assert config.max_poll == send.UdpIbvConfig.DEFAULT_MAX_POLL
        assert config.memory_regions == []

    def test_kwargs_construct(self):
        data1 = bytearray(10)
        data2 = bytearray(10)
        config = send.UdpIbvConfig(
            endpoints=[('hello', 1234), ('goodbye', 2345)],
            interface_address='1.2.3.4',
            buffer_size=100,
            ttl=2,
            comp_vector=-1,
            max_poll=1000,
            memory_regions=[data1, data2])
        assert config.endpoints == [('hello', 1234), ('goodbye', 2345)]
        assert config.interface_address == '1.2.3.4'
        assert config.buffer_size == 100
        assert config.ttl == 2
        assert config.comp_vector == -1
        assert config.max_poll == 1000
        assert config.memory_regions == [data1, data2]

    def test_default_buffer_size(self):
        config = send.UdpIbvConfig()
        config.buffer_size = 0
        assert config.buffer_size == send.UdpIbvConfig.DEFAULT_BUFFER_SIZE

    def test_bad_max_poll(self):
        config = send.UdpIbvConfig()
        with pytest.raises(ValueError):
            config.max_poll = 0
        with pytest.raises(ValueError):
            config.max_poll = -1

    def test_empty_memory_region(self):
        data = bytearray(0)
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='10.0.0.1',
            memory_regions=[data])
        with pytest.raises(ValueError, match='memory region must have non-zero size'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_overlapping_memory_regions(self):
        data = memoryview(bytearray(10))
        part1 = data[:6]
        part2 = data[5:]
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='10.0.0.1',
            memory_regions=[part1, part2])
        with pytest.raises(ValueError, match='memory regions overlap'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_no_endpoints(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(interface_address='10.0.0.1')
        with pytest.raises(ValueError, match='endpoints is empty'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_ipv6_endpoints(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('::1', 8888)],
            interface_address='10.0.0.1')
        with pytest.raises(ValueError, match='endpoint is not an IPv4 multicast address'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_unicast_endpoints(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('10.0.0.1', 8888)],
            interface_address='10.0.0.1')
        with pytest.raises(ValueError, match='endpoint is not an IPv4 multicast address'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_no_interface_address(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)])
        with pytest.raises(ValueError, match='interface address'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_bad_interface_address(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='this is not an interface address')
        with pytest.raises(RuntimeError, match='Host not found'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_ipv6_interface_address(self):
        config = send.StreamConfig()
        udp_ibv_config = send.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='::1')
        with pytest.raises(ValueError, match='interface address'):
            send.UdpIbvStream(spead2.ThreadPool(), config, udp_ibv_config)

    def test_deprecated_constants(self):
        with pytest.deprecated_call():
            assert send.UdpIbvStream.DEFAULT_BUFFER_SIZE == send.UdpIbvConfig.DEFAULT_BUFFER_SIZE
        with pytest.deprecated_call():
            assert send.UdpIbvStream.DEFAULT_MAX_POLL == send.UdpIbvConfig.DEFAULT_MAX_POLL



# Copyright 2018-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test that data can be passed through the various async transports."""

import socket
import asyncio

import pytest

import spead2
import spead2.send
import spead2.recv.asyncio
import spead2.send.asyncio

from . import test_passthrough


class BaseTestPassthroughAsync(test_passthrough.BaseTestPassthrough):
    def transmit_item_groups(self, item_groups, memcpy, allocator, new_order='='):
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(
                self.transmit_item_groups_async(item_groups, memcpy, allocator, new_order))
        finally:
            loop.close()

    async def transmit_item_groups_async(self, item_groups, memcpy, allocator, new_order='='):
        if self.requires_ipv6:
            self.check_ipv6()
        recv_config = spead2.recv.StreamConfig(memcpy=memcpy)
        if allocator is not None:
            recv_config.memory_allocator = allocator
        receivers = [
            spead2.recv.asyncio.Stream(spead2.ThreadPool(), recv_config)
            for i in range(len(item_groups))
        ]
        await self.prepare_receivers(receivers)
        sender = await self.prepare_senders(spead2.ThreadPool(), len(item_groups))
        gens = [
            spead2.send.HeapGenerator(item_group)
            for item_group in item_groups
        ]
        if len(item_groups) != 1:
            # Use reversed order so that if everything is actually going
            # through the same transport it will get picked up.
            for i, gen in reversed(list(enumerate(gens))):
                await sender.async_send_heap(gen.get_heap(), substream_index=i)
            for i, gen in enumerate(gens):
                await sender.async_send_heap(gen.get_end(), substream_index=i)
        else:
            # This is a separate code path to give coverage of the case where
            # the substream index is implicit.
            await sender.async_send_heap(gens[0].get_heap())
            await sender.async_send_heap(gens[0].get_end())
        await sender.async_flush()
        received_item_groups = []
        for receiver in receivers:
            ig = spead2.ItemGroup()
            for heap in receiver:
                ig.update(heap, new_order)
            received_item_groups.append(ig)
        return received_item_groups

    async def prepare_receivers(self, receivers):
        """Generate receivers to use in the test."""
        assert len(receivers) == 1
        await self.prepare_receiver(receivers[0])

    async def prepare_senders(self, thread_pool, n):
        """Generate a sender to use in the test, with `n` substreams."""
        assert n == 1
        return await self.prepare_sender(thread_pool)


class BaseTestPassthroughSubstreamsAsync(
        test_passthrough.BaseTestPassthroughSubstreams, BaseTestPassthroughAsync):
    async def prepare_receivers(self, receivers):
        raise NotImplementedError()

    async def prepare_senders(self, thread_pool, n):
        raise NotImplementedError()


class TestPassthroughUdp(BaseTestPassthroughSubstreamsAsync):
    async def prepare_receivers(self, receivers):
        for i, receiver in enumerate(receivers):
            receiver.add_udp_reader(8888 + i, bind_hostname="localhost")

    async def prepare_senders(self, thread_pool, n):
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.asyncio.UdpStream(
                    thread_pool, "localhost", 8888,
                    spead2.send.StreamConfig(rate=1e7),
                    buffer_size=0)
        else:
            return spead2.send.asyncio.UdpStream(
                thread_pool,
                [("localhost", 8888 + i) for i in range(n)],
                spead2.send.StreamConfig(rate=1e7),
                buffer_size=0)


class TestPassthroughUdpCustomSocket(BaseTestPassthroughSubstreamsAsync):
    async def prepare_receivers(self, receivers):
        self._ports = []
        for receiver in receivers:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.bind(('127.0.0.1', 0))
            self._ports.append(sock.getsockname()[1])
            receiver.add_udp_reader(sock)
            sock.close()

    async def prepare_senders(self, thread_pool, n):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        if n == 1:
            with pytest.deprecated_call():
                stream = spead2.send.asyncio.UdpStream(
                    thread_pool, sock, '127.0.0.1', self._ports[0],
                    spead2.send.StreamConfig(rate=1e7))
        else:
            stream = spead2.send.asyncio.UdpStream(
                thread_pool, sock,
                [('127.0.0.1', port) for port in self._ports],
                spead2.send.StreamConfig(rate=1e7))
        sock.close()
        return stream


class TestPassthroughTcp(BaseTestPassthroughAsync):
    async def prepare_receiver(self, receiver):
        receiver.add_tcp_reader(8888, bind_hostname="127.0.0.1")

    async def prepare_sender(self, thread_pool):
        sender = await spead2.send.asyncio.TcpStream.connect(
            thread_pool, [("127.0.0.1", 8888)])
        return sender


class TestPassthroughTcpCustomSocket(BaseTestPassthroughAsync):
    async def prepare_receiver(self, receiver):
        sock = socket.socket()
        # Prevent second iteration of the test from failing
        sock.bind(('127.0.0.1', 0))
        self._port = sock.getsockname()[1]
        sock.listen(1)
        receiver.add_tcp_reader(sock)
        sock.close()

    async def prepare_sender(self, thread_pool):
        sock = socket.socket()
        sock.setblocking(False)
        await asyncio.get_event_loop().sock_connect(sock, ('127.0.0.1', self._port))
        sender = spead2.send.asyncio.TcpStream(
            thread_pool, sock)
        sock.close()
        return sender


class TestPassthroughInproc(BaseTestPassthroughSubstreamsAsync):
    async def prepare_receivers(self, receivers):
        assert len(receivers) == len(self._queues)
        for receiver, queue in zip(receivers, self._queues):
            receiver.add_inproc_reader(queue)

    async def prepare_senders(self, thread_pool, n):
        assert n == len(self._queues)
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.asyncio.InprocStream(thread_pool, self._queues[0])
        else:
            return spead2.send.asyncio.InprocStream(thread_pool, self._queues)

    async def transmit_item_groups_async(self, item_groups, memcpy, allocator, new_order='='):
        self._queues = [spead2.InprocQueue() for ig in item_groups]
        ret = await super().transmit_item_groups_async(
            item_groups, memcpy, allocator, new_order)
        for queue in self._queues:
            queue.stop()
        return ret



# Copyright 2019 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import AsyncIterator, Optional

import spead2
import spead2.recv

class Stream(spead2.recv._Stream):
    async def get(self) -> spead2.recv.Heap: ...
    def __aiter__(self) -> AsyncIterator[spead2.recv.Heap]: ...



# Copyright 2015, 2017, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket
import struct
import time

import numpy as np
import pytest

import spead2
import spead2.recv as recv
import spead2.send as send


class Item:
    def __init__(self, id, value, immediate=False, offset=0):
        self.id = id
        self.value = value
        self.immediate = immediate
        self.offset = offset

    def encode(self, heap_address_bits):
        if self.immediate:
            return struct.pack('>Q', (1 << 63) | (self.id << heap_address_bits) | self.value)
        else:
            return struct.pack('>Q', (self.id << heap_address_bits) | self.offset)


class Flavour:
    def __init__(self, heap_address_bits, bug_compat=0):
        self.heap_address_bits = heap_address_bits
        self.bug_compat = bug_compat

    def _encode_be(self, size, value):
        """Encodes `value` as big-endian in `size` bytes"""
        assert size <= 8
        packed = struct.pack('>Q', value)
        return packed[8 - size:]

    def make_packet(self, items, payload):
        """Generate data for a packet at a low level. The value of non-immediate
        items are taken from the payload; the values in the objects are ignored.
        """
        data = []
        data.append(struct.pack('>BBBBHH', 0x53, 0x4,
                                (64 - self.heap_address_bits) // 8,
                                self.heap_address_bits // 8, 0,
                                len(items)))
        for item in items:
            data.append(item.encode(self.heap_address_bits))
        data.append(bytes(payload))
        return b''.join(data)

    def make_format(self, format):
        if self.bug_compat & spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS:
            field_size = 7
        else:
            field_size = (64 - self.heap_address_bits) // 8
        data = []
        for field in format:
            assert len(field[0]) == 1
            data.append(field[0].encode('ascii'))
            data.append(self._encode_be(field_size, field[1]))
        return b''.join(data)

    def make_shape(self, shape):
        if self.bug_compat & spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS:
            field_size = 8
        else:
            field_size = self.heap_address_bits // 8 + 1
        if self.bug_compat & spead2.BUG_COMPAT_SHAPE_BIT_1:
            variable_marker = 2
        else:
            variable_marker = 1

        data = []
        for value in shape:
            if value is None:
                data.append(struct.pack('>B', variable_marker))
                data.append(self._encode_be(field_size - 1, 0))
            elif value >= 0:
                data.append(self._encode_be(field_size, value))
            else:
                raise ValueError('Shape must contain non-negative values and None')
        return b''.join(data)

    def make_packet_heap(self, heap_cnt, items):
        """Construct a single-packet heap."""
        payload_size = 0
        for item in items:
            if not item.immediate:
                payload_size += len(bytes(item.value))

        all_items = [
            Item(spead2.HEAP_CNT_ID, heap_cnt, True),
            Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
            Item(spead2.PAYLOAD_LENGTH_ID, payload_size, True),
            Item(spead2.HEAP_LENGTH_ID, payload_size, True)]
        offset = 0
        payload = bytearray(payload_size)
        for item in items:
            if not item.immediate:
                value = bytes(item.value)
                all_items.append(Item(item.id, value, offset=offset))
                payload[offset : offset + len(value)] = value
                offset += len(value)
            else:
                all_items.append(item)
        return self.make_packet(all_items, payload)

    def make_plain_descriptor(self, id, name, description, format, shape):
        if not isinstance(name, bytes):
            name = name.encode('ascii')
        if not isinstance(description, bytes):
            description = description.encode('ascii')
        return Item(spead2.DESCRIPTOR_ID, self.make_packet_heap(
            1,
            [
                Item(spead2.DESCRIPTOR_ID_ID, id, True),
                Item(spead2.DESCRIPTOR_NAME_ID, name),
                Item(spead2.DESCRIPTOR_DESCRIPTION_ID, description),
                Item(spead2.DESCRIPTOR_FORMAT_ID, self.make_format(format)),
                Item(spead2.DESCRIPTOR_SHAPE_ID, self.make_shape(shape))
            ]))

    def make_numpy_descriptor_raw(self, id, name, description, header):
        if not isinstance(name, bytes):
            name = name.encode('ascii')
        if not isinstance(description, bytes):
            description = description.encode('ascii')
        if not isinstance(header, bytes):
            header = header.encode('ascii')
        return Item(spead2.DESCRIPTOR_ID, self.make_packet_heap(
            1,
            [
                Item(spead2.DESCRIPTOR_ID_ID, id, True),
                Item(spead2.DESCRIPTOR_NAME_ID, name),
                Item(spead2.DESCRIPTOR_DESCRIPTION_ID, description),
                Item(spead2.DESCRIPTOR_DTYPE_ID, header)
            ]))

    def make_numpy_descriptor(self, id, name, description, dtype, shape, fortran_order=False):
        header = str({
            'descr': np.lib.format.dtype_to_descr(np.dtype(dtype)),
            'fortran_order': bool(fortran_order),
            'shape': tuple(shape)
        })
        return self.make_numpy_descriptor_raw(id, name, description, header)

    def make_numpy_descriptor_from(self, id, name, description, array):
        if array.flags.c_contiguous:
            fortran_order = False
        elif array.flags.f_contiguous:
            fortran_order = True
        else:
            raise ValueError('Array must be C or Fortran-order contiguous')
        return self.make_numpy_descriptor(
            id, name, description, array.dtype, array.shape, fortran_order)


FLAVOUR = Flavour(48)


class TestDecode:
    """Various types of descriptors must be correctly interpreted to decode data"""

    def setup(self):
        self.flavour = FLAVOUR

    def data_to_heaps(self, data, **kwargs):
        """Take some data and pass it through the receiver to obtain a set of heaps.

        Keyword arguments are passed to the receiver constructor.
        """
        thread_pool = spead2.ThreadPool()
        config = recv.StreamConfig(bug_compat=self.flavour.bug_compat)
        ring_config = recv.RingStreamConfig()
        for key in ['stop_on_stop_item', 'allow_unsized_heaps', 'allow_out_of_order']:
            if key in kwargs:
                setattr(config, key, kwargs.pop(key))
        for key in ['contiguous_only', 'incomplete_keep_payload_ranges']:
            if key in kwargs:
                setattr(ring_config, key, kwargs.pop(key))
        if kwargs:
            raise ValueError(f'Unexpected keyword arguments ({kwargs})')
        stream = recv.Stream(thread_pool, config, ring_config)
        stream.add_buffer_reader(data)
        return list(stream)

    def data_to_ig(self, data):
        """Take some data and pass it through the receiver to obtain a single heap,
        from which the items are extracted.
        """
        heaps = self.data_to_heaps(data)
        assert len(heaps) == 1
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        for name, item in ig.items():
            assert item.name == name
        return ig

    def data_to_item(self, data, expected_id):
        """Take some data and pass it through the receiver to obtain a single heap,
        with a single item, which is returned.
        """
        ig = self.data_to_ig(data)
        assert len(ig) == 1
        assert expected_id in ig
        return ig[expected_id]

    def test_scalar_int(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('i', 32)], []),
                Item(0x1234, struct.pack('>i', -123456789))
            ])
        item = self.data_to_item(packet, 0x1234)
        assert isinstance(item.value, np.int32)
        assert item.value == -123456789

    def test_scalar_int_immediate(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('u', 32)], []),
                Item(0x1234, 0x12345678, True)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert isinstance(item.value, np.uint32)
        assert item.value == 0x12345678

    def test_scalar_int_immediate_fastpath(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_uint', 'a scalar integer', [('u', 48)], []),
                self.flavour.make_plain_descriptor(
                    0x1235, 'test_scalar_positive', 'a positive scalar integer', [('i', 48)], []),
                self.flavour.make_plain_descriptor(
                    0x1236, 'test_scalar_negative', 'a negative scalar integer', [('i', 48)], []),
                Item(0x1234, 0x9234567890AB, True),
                Item(0x1235, 0x1234567890AB, True),
                Item(0x1236, 0x9234567890AB, True)
            ])
        ig = self.data_to_ig(packet)
        assert len(ig) == 3
        assert ig[0x1234].value == 0x9234567890AB
        assert ig[0x1235].value == 0x1234567890AB
        assert ig[0x1236].value == -0x6DCBA9876F55

    def test_string(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, b'Hello world')
            ])
        item = self.data_to_item(packet, 0x1234)
        assert item.value == 'Hello world'

    def test_array(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_array', 'an array of floats', [('f', 32)], (3, 2)),
                Item(0x1234, struct.pack('>6f', *np.arange(1.5, 7.5)))
            ])
        item = self.data_to_item(packet, 0x1234)
        expected = np.array([[1.5, 2.5], [3.5, 4.5], [5.5, 6.5]], dtype=np.float32)
        np.testing.assert_equal(expected, item.value)

    def test_array_fields(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_array', 'an array of floats', [('f', 32), ('i', 8)], (3,)),
                Item(0x1234, struct.pack('>fbfbfb', 1.5, 1, 2.5, 2, 4.5, -4))
            ])
        item = self.data_to_item(packet, 0x1234)
        dtype = np.dtype('=f4,i1')
        assert item.value.dtype == dtype
        expected = np.array([(1.5, 1), (2.5, 2), (4.5, -4)], dtype=dtype)
        np.testing.assert_equal(expected, item.value)

    def test_array_numpy(self):
        expected = np.array([[1.25, 1.75], [2.25, 2.5]], dtype=np.float64)
        send = expected
        if self.flavour.bug_compat & spead2.BUG_COMPAT_SWAP_ENDIAN:
            send = expected.byteswap()
        else:
            send = expected
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor_from(
                    0x1234, 'test_array_numpy', 'an array of floats', expected),
                Item(0x1234, send.data)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert item.value.dtype == expected.dtype
        np.testing.assert_equal(expected, item.value)

    def test_array_numpy_fortran(self):
        expected = np.array([[1.25, 1.75], [2.25, 2.5]], dtype=np.float64).T
        assert not expected.flags.c_contiguous
        send = expected
        if self.flavour.bug_compat & spead2.BUG_COMPAT_SWAP_ENDIAN:
            send = expected.byteswap()
        else:
            send = expected
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor_from(
                    0x1234, 'test_array_numpy_fortran', 'an array of floats', expected),
                Item(0x1234, np.ravel(send, 'K').data)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert item.value.dtype == expected.dtype
        np.testing.assert_equal(expected, item.value)

    def test_fallback_uint(self):
        expected = np.array([0xABC, 0xDEF, 0x123])
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array of 12-bit uints', [('u', 12)], (3,)),
                Item(0x1234, b'\xAB\xCD\xEF\x12\x30')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_int(self):
        expected = np.array([-1348, -529, 291])
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array of 12-bit ints', [('i', 12)], (3,)),
                Item(0x1234, b'\xAB\xCD\xEF\x12\x30')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_types(self):
        expected = np.array([(True, 17, 'y', 1.0), (False, -23, 'n', -1.0)], dtype='O,O,S1,>f4')
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array with bools, int, chars and floats',
                    [('b', 1), ('i', 7), ('c', 8), ('f', 32)], (2,)),
                Item(0x1234, b'\x91y\x3F\x80\x00\x00' + b'\x69n\xBF\x80\x00\x00')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_scalar(self):
        expected = 0x1234567890AB
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_scalar', 'a scalar with unusual type',
                    [('u', 48)], ()),
                Item(0x1234, b'\x12\x34\x56\x78\x90\xAB')
            ])
        item = self.data_to_item(packet, 0x1234)
        assert item.value == expected

    def test_duplicate_item_pointers(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[0 : 32]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[32 : 64])
        ]
        heaps = self.data_to_heaps(b''.join(packets))
        assert len(heaps) == 1
        items = heaps[0].get_items()
        assert len(items) == 2
        items.sort(key=lambda item: item.id)
        assert items[0].id == 0x1600
        assert items[0].is_immediate is True
        assert items[0].immediate_value == 12345
        assert items[1].id == 0x5000
        assert items[1].is_immediate is False
        assert bytearray(items[1]) == payload

    def test_out_of_order_packets(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[32 : 64]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True)], payload[0 : 32])
        ]
        heaps = self.data_to_heaps(b''.join(packets), allow_out_of_order=True)
        assert len(heaps) == 1
        items = heaps[0].get_items()
        assert len(items) == 2
        items.sort(key=lambda item: item.id)
        assert items[0].id == 0x1600
        assert items[0].is_immediate is True
        assert items[0].immediate_value == 12345
        assert items[1].id == 0x5000
        assert items[1].is_immediate is False
        assert bytearray(items[1]) == payload

    def test_out_of_order_disallowed(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[32 : 64]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True)], payload[0 : 32])
        ]
        heaps = self.data_to_heaps(b''.join(packets),
                                   contiguous_only=False,
                                   incomplete_keep_payload_ranges=True)
        assert len(heaps) == 1
        items = heaps[0].get_items()
        assert items == []
        assert heaps[0].heap_length == 64
        assert heaps[0].received_length == 32
        assert heaps[0].payload_ranges == [(0, 32)]

    def test_incomplete_heaps(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 5, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 7, True),
                Item(spead2.HEAP_LENGTH_ID, 96, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[5 : 12]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 96, True)], payload[32 : 64])
        ]
        heaps = self.data_to_heaps(b''.join(packets),
                                   contiguous_only=False,
                                   incomplete_keep_payload_ranges=True,
                                   allow_out_of_order=True)
        assert len(heaps) == 1
        assert isinstance(heaps[0], recv.IncompleteHeap)
        items = heaps[0].get_items()
        assert len(items) == 1        # Addressed item must be excluded
        assert items[0].id == 0x1600
        assert heaps[0].heap_length == 96
        assert heaps[0].received_length == 39
        assert heaps[0].payload_ranges == [(5, 12), (32, 64)]

    def test_is_start_of_stream(self):
        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_START, immediate=True)])
        heaps = self.data_to_heaps(packet)
        assert heaps[0].is_start_of_stream() is True

        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_DESCRIPTOR_REISSUE, immediate=True)])
        heaps = self.data_to_heaps(packet)
        assert heaps[0].is_start_of_stream() is False

    def test_is_end_of_stream(self):
        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_STOP, immediate=True)])
        heaps = self.data_to_heaps(packet, stop_on_stop_item=False)
        assert heaps[0].is_end_of_stream() is True
        assert heaps[0].is_start_of_stream() is False

    def test_no_stop_on_stop_item(self):
        packet1 = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_STOP, immediate=True)])
        packet2 = self.flavour.make_packet_heap(
            2,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, b'Hello world')
            ])
        heaps = self.data_to_heaps(packet1 + packet2, stop_on_stop_item=False)
        assert len(heaps) == 2
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        ig.update(heaps[1])
        assert ig['test_string'].value == 'Hello world'

    def test_size_mismatch(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'bad', 'an item with insufficient data', [('u', 32)], (5, 5)),
                Item(0x1234, b'\0' * 99)
            ])
        heaps = self.data_to_heaps(packet)
        assert len(heaps) == 1
        ig = spead2.ItemGroup()
        with pytest.raises(ValueError):
            ig.update(heaps[0])

    def test_numpy_object(self):
        """numpy dtypes can contain Python objects (by pointer). These can't be
        used for SPEAD.
        """
        dtype = np.dtype('f4,O')
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor(
                    0x1234, 'object', 'an item with object pointers', dtype, (5,)),
                Item(0x1234, b'?' * 100)
            ])
        heaps = self.data_to_heaps(packet)
        assert len(heaps) == 1
        ig = spead2.ItemGroup()
        with pytest.raises(ValueError):
            ig.update(heaps[0])

    def test_numpy_zero_size(self):
        """numpy dtypes can represent zero bytes."""
        dtype = np.dtype(np.str_)
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor(
                    0x1234, 'empty', 'an item with zero-byte dtype', dtype, (5,)),
                Item(0x1234, b'')
            ])
        heaps = self.data_to_heaps(packet)
        assert len(heaps) == 1
        ig = spead2.ItemGroup()
        with pytest.raises(ValueError):
            ig.update(heaps[0])

    def test_numpy_malformed(self):
        """Malformed numpy header must raise :py:exc:`ValueError`."""
        def helper(header):
            packet = self.flavour.make_packet_heap(
                1,
                [
                    self.flavour.make_numpy_descriptor_raw(
                        0x1234, 'name', 'description', header)
                ])
            heaps = self.data_to_heaps(packet)
            assert len(heaps) == 1
            ig = spead2.ItemGroup()
            with pytest.raises(ValueError):
                ig.update(heaps[0])
        helper("{'descr': 'S1'")   # Syntax error: no closing brace
        helper("123")              # Not a dictionary
        helper("import os")        # Security check
        helper("{'descr': 'S1'}")  # Missing keys
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (), 'foo': 'bar'}")  # Extra keys
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (-1,)}")    # Bad shape
        helper("{'descr': 1, 'fortran_order': False, 'shape': ()}")          # Bad descriptor
        helper("{'descr': '+-', 'fortran_order': False, 'shape': ()}")       # Bad descriptor
        helper("{'descr': 'S1', 'fortran_order': 0, 'shape': ()}")           # Bad fortran_order
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (None,)}")  # Bad shape

    def test_nonascii_value(self):
        """Receiving non-ASCII characters in a c8 string must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, '\u0200'.encode())
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        with pytest.raises(UnicodeDecodeError):
            ig.update(heaps[0])

    def test_nonascii_name(self):
        """Receiving non-ASCII characters in an item name must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, b'\xEF', 'a byte string', [('c', 8)], [None])
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        with pytest.raises(UnicodeDecodeError):
            ig.update(heaps[0])

    def test_nonascii_description(self):
        """Receiving non-ASCII characters in an item description must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'name', b'\xEF', [('c', 8)], [None])
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        with pytest.raises(UnicodeDecodeError):
            ig.update(heaps[0])

    def test_no_heap_size(self):
        """Heap consisting of packets with no heap size items must work.
        This also tests mixing the general items in with the packet items.
        """
        payload1 = bytes(np.arange(0, 64, dtype=np.uint8).data)
        payload2 = bytes(np.arange(64, 96, dtype=np.uint8).data)
        packet1 = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(0x1000, None, False, offset=0),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True)
            ], payload1)
        packet2 = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 64, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True)
            ], payload2)
        heaps = self.data_to_heaps(packet1 + packet2)
        assert len(heaps) == 1
        raw_items = heaps[0].get_items()
        assert len(raw_items) == 1
        assert bytearray(raw_items[0]) == payload1 + payload2

    def test_disallow_unsized_heaps(self, caplog):
        """Packets without heap length rejected if disallowed"""
        packet = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(0x1000, None, False, offset=0),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True)
            ], bytes(np.arange(0, 64, dtype=np.uint8).data))
        with caplog.at_level('INFO', 'spead2'):
            heaps = self.data_to_heaps(packet, allow_unsized_heaps=False)
            # Logging is asynchronous, so we have to give it a bit of time
            time.sleep(0.1)
        assert caplog.messages == ['packet rejected because it has no HEAP_LEN']
        assert len(heaps) == 0

    def test_bad_offset(self):
        """Heap with out-of-range offset should be dropped"""
        packet = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True),
                Item(0x1000, None, False, offset=65)
            ], b'\0' * 64)
        heaps = self.data_to_heaps(packet)
        assert len(heaps) == 0


class TestStreamConfig:
    """Tests for :class:`spead2.recv.StreamConfig`."""

    def test_default_construct(self):
        config = recv.StreamConfig()
        assert config.max_heaps == recv.StreamConfig.DEFAULT_MAX_HEAPS
        assert config.bug_compat == 0
        assert config.memcpy == spead2.MEMCPY_STD
        assert config.stop_on_stop_item is True
        assert config.allow_unsized_heaps is True
        assert config.allow_out_of_order is False

    def test_set_get(self):
        config = recv.StreamConfig()
        config.max_heaps = 5
        config.bug_compat = spead2.BUG_COMPAT_PYSPEAD_0_5_2
        config.memcpy = spead2.MEMCPY_NONTEMPORAL
        config.stop_on_stop_item = False
        config.allow_unsized_heaps = False
        config.allow_out_of_order = True
        config.memory_allocator = allocator = spead2.MmapAllocator()
        assert config.max_heaps == 5
        assert config.bug_compat == spead2.BUG_COMPAT_PYSPEAD_0_5_2
        assert config.memcpy == spead2.MEMCPY_NONTEMPORAL
        assert config.memory_allocator is allocator
        assert config.stop_on_stop_item is False
        assert config.allow_unsized_heaps is False
        assert config.allow_out_of_order is True

    def test_kwargs_construct(self):
        config = recv.StreamConfig(
            max_heaps=5,
            bug_compat=spead2.BUG_COMPAT_PYSPEAD_0_5_2,
            memcpy=spead2.MEMCPY_NONTEMPORAL,
            stop_on_stop_item=False,
            allow_unsized_heaps=False,
            allow_out_of_order=True
        )
        assert config.max_heaps == 5
        assert config.bug_compat == spead2.BUG_COMPAT_PYSPEAD_0_5_2
        assert config.memcpy == spead2.MEMCPY_NONTEMPORAL
        assert config.stop_on_stop_item is False
        assert config.allow_unsized_heaps is False
        assert config.allow_out_of_order is True

    def test_max_heaps_zero(self):
        """Constructing a config with max_heaps=0 raises ValueError"""
        with pytest.raises(ValueError):
            recv.StreamConfig(max_heaps=0)

    def test_bad_bug_compat(self):
        with pytest.raises(ValueError):
            recv.StreamConfig(bug_compat=0xff)

    def test_bad_kwarg(self):
        with pytest.raises(TypeError):
            recv.StreamConfig(not_valid_arg=1)


class TestRingStreamConfig:
    """Tests for :class:`spead2.recv.StreamConfig`."""

    def test_default_construct(self):
        config = recv.RingStreamConfig()
        assert config.heaps == recv.RingStreamConfig.DEFAULT_HEAPS
        assert config.contiguous_only is True
        assert config.incomplete_keep_payload_ranges is False

    def test_set_get(self):
        config = recv.RingStreamConfig()
        config.heaps = 5
        config.contiguous_only = False
        config.incomplete_keep_payload_ranges = True
        assert config.heaps == 5
        assert config.contiguous_only is False
        assert config.incomplete_keep_payload_ranges is True

    def test_heaps_zero(self):
        """Constructing a config with heaps=0 raises ValueError"""
        with pytest.raises(ValueError):
            recv.RingStreamConfig(heaps=0)


@pytest.mark.skipif(not hasattr(spead2, 'IbvContext'), reason='IBV support not compiled in')
class TestUdpIbvConfig:
    def test_default_construct(self):
        config = recv.UdpIbvConfig()
        assert config.endpoints == []
        assert config.interface_address == ''
        assert config.buffer_size == recv.UdpIbvConfig.DEFAULT_BUFFER_SIZE
        assert config.max_size == recv.UdpIbvConfig.DEFAULT_MAX_SIZE
        assert config.comp_vector == 0
        assert config.max_poll == spead2.send.UdpIbvConfig.DEFAULT_MAX_POLL

    def test_kwargs_construct(self):
        config = recv.UdpIbvConfig(
            endpoints=[('hello', 1234), ('goodbye', 2345)],
            interface_address='1.2.3.4',
            buffer_size=100,
            max_size=4321,
            comp_vector=-1,
            max_poll=1000)
        assert config.endpoints == [('hello', 1234), ('goodbye', 2345)]
        assert config.interface_address == '1.2.3.4'
        assert config.buffer_size == 100
        assert config.max_size == 4321
        assert config.comp_vector == -1
        assert config.max_poll == 1000

    def test_default_buffer_size(self):
        config = recv.UdpIbvConfig()
        config.buffer_size = 0
        assert config.buffer_size == recv.UdpIbvConfig.DEFAULT_BUFFER_SIZE

    def test_bad_max_poll(self):
        config = recv.UdpIbvConfig()
        with pytest.raises(ValueError):
            config.max_poll = 0
        with pytest.raises(ValueError):
            config.max_poll = -1

    def test_bad_max_size(self):
        config = recv.UdpIbvConfig()
        with pytest.raises(ValueError):
            config.max_size = 0

    def test_no_endpoints(self):
        config = recv.UdpIbvConfig(interface_address='10.0.0.1')
        stream = recv.Stream(
            spead2.ThreadPool(), recv.StreamConfig(), recv.RingStreamConfig())
        with pytest.raises(ValueError, match='endpoints is empty'):
            stream.add_udp_ibv_reader(config)

    def test_ipv6_endpoints(self):
        config = recv.UdpIbvConfig(
            endpoints=[('::1', 8888)],
            interface_address='10.0.0.1')
        stream = recv.Stream(
            spead2.ThreadPool(), recv.StreamConfig(), recv.RingStreamConfig())
        with pytest.raises(ValueError, match='endpoint is not an IPv4 address'):
            stream.add_udp_ibv_reader(config)

    def test_no_interface_address(self):
        config = recv.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)])
        stream = recv.Stream(
            spead2.ThreadPool(), recv.StreamConfig(), recv.RingStreamConfig())
        with pytest.raises(ValueError, match='interface address'):
            stream.add_udp_ibv_reader(config)

    def test_bad_interface_address(self):
        config = recv.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='this is not an interface address')
        stream = recv.Stream(
            spead2.ThreadPool(), recv.StreamConfig(), recv.RingStreamConfig())
        with pytest.raises(RuntimeError, match='Host not found'):
            stream.add_udp_ibv_reader(config)

    def test_ipv6_interface_address(self):
        config = recv.UdpIbvConfig(
            endpoints=[('239.255.88.88', 8888)],
            interface_address='::1')
        stream = recv.Stream(
            spead2.ThreadPool(), recv.StreamConfig(), recv.RingStreamConfig())
        with pytest.raises(ValueError, match='interface address'):
            stream.add_udp_ibv_reader(config)

    def test_deprecated_constants(self):
        with pytest.deprecated_call():
            assert recv.Stream.DEFAULT_UDP_IBV_BUFFER_SIZE == recv.UdpIbvConfig.DEFAULT_BUFFER_SIZE
        with pytest.deprecated_call():
            assert recv.Stream.DEFAULT_UDP_IBV_MAX_SIZE == recv.UdpIbvConfig.DEFAULT_MAX_SIZE
        with pytest.deprecated_call():
            assert recv.Stream.DEFAULT_UDP_IBV_MAX_POLL == recv.UdpIbvConfig.DEFAULT_MAX_POLL


class TestStream:
    """Tests for the stream API."""

    def setup(self):
        self.flavour = FLAVOUR

    def test_full_stop(self):
        """Must be able to stop even if the consumer is not consuming
        anything."""
        thread_pool = spead2.ThreadPool(1)
        sender = send.BytesStream(thread_pool)
        ig = send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        gen = send.HeapGenerator(ig)
        for i in range(10):
            sender.send_heap(gen.get_heap(data='all'))
        recv_ring_config = recv.RingStreamConfig(heaps=4)
        receiver = recv.Stream(thread_pool, ring_config=recv_ring_config)
        receiver.add_buffer_reader(sender.getvalue())
        # Wait for the ring buffer to block
        while receiver.stats.worker_blocked == 0:
            time.sleep(0.0)
        # Can't usefully check the stats here, because they're only
        # updated at the end of a batch.
        assert receiver.ringbuffer.capacity() == 4
        assert receiver.ringbuffer.size() == 4

        receiver.stop()            # This unblocks all remaining heaps
        stats = receiver.stats
        assert stats.heaps == 10
        assert stats.packets == 10
        assert stats.incomplete_heaps_evicted == 0
        assert stats.incomplete_heaps_flushed == 0
        assert stats.worker_blocked == 1
        assert receiver.ringbuffer.capacity() == 4
        assert receiver.ringbuffer.size() == 4

    def test_no_stop_heap(self):
        """A heap containing a stop is not passed to the ring"""
        thread_pool = spead2.ThreadPool(1)
        sender = send.BytesStream(thread_pool)
        ig = send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        gen = send.HeapGenerator(ig)
        sender.send_heap(gen.get_heap(data='all'))
        sender.send_heap(gen.get_end())
        receiver = recv.Stream(thread_pool)
        receiver.add_buffer_reader(sender.getvalue())
        heaps = list(receiver)
        assert len(heaps) == 1

        stats = receiver.stats
        assert stats.heaps == 1
        assert stats.packets == 2
        assert stats.incomplete_heaps_evicted == 0
        assert stats.incomplete_heaps_flushed == 0
        assert stats.worker_blocked == 0


class TestUdpReader:
    def test_out_of_range_udp_port(self):
        receiver = recv.Stream(spead2.ThreadPool())
        with pytest.raises(TypeError):
            receiver.add_udp_reader(100000)

    def test_illegal_udp_port(self):
        receiver = recv.Stream(spead2.ThreadPool())
        with pytest.raises(RuntimeError):
            receiver.add_udp_reader(22)


class TestTcpReader:
    def setup(self):
        self.receiver = recv.Stream(spead2.ThreadPool())
        recv_sock = socket.socket()
        recv_sock.bind(("127.0.0.1", 0))
        recv_sock.listen(1)
        port = recv_sock.getsockname()[1]
        self.receiver.add_tcp_reader(acceptor=recv_sock)
        recv_sock.close()

        self.send_sock = socket.socket()
        self.send_sock.connect(("127.0.0.1", port))

    def teardown(self):
        self.close()

    def close(self):
        if self.send_sock is not None:
            self.send_sock.close()
            self.send_sock = None

    def data_to_heaps(self, data):
        """Take some data and pass it through the receiver to obtain a set of heaps.

        The send socket is closed.
        """
        self.send_sock.send(data)
        self.close()
        return list(self.receiver)

    def data_to_ig(self, data):
        """Take some data and pass it through the receiver to obtain a single heap,
        from which the items are extracted.
        """
        heaps = self.data_to_heaps(data)
        assert len(heaps) == 1
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        for name, item in ig.items():
            assert item.name == name
        return ig

    def data_to_item(self, data, expected_id):
        """Take some data and pass it through the receiver to obtain a single heap,
        with a single item, which is returned.
        """
        ig = self.data_to_ig(data)
        assert len(ig) == 1
        assert expected_id in ig
        return ig[expected_id]

    def simple_packet(self):
        return FLAVOUR.make_packet_heap(
            1,
            [
                FLAVOUR.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('u', 32)], []),
                Item(0x1234, 0x12345678, True)
            ])

    def test_bad_header(self):
        """A nonsense header followed by a normal packet"""
        data = b'deadbeef' + self.simple_packet()
        item = self.data_to_item(data, 0x1234)
        assert isinstance(item.value, np.uint32)
        assert item.value == 0x12345678

    def test_packet_too_big(self):
        """A packet that is too large (should be rejected) followed by a normal packet"""
        zeros = np.zeros(100000, np.uint8)
        packet1 = FLAVOUR.make_packet_heap(
            1,
            [
                FLAVOUR.make_numpy_descriptor_from(
                    0x2345, 'test_big_array', 'over-sized packet', zeros),
                Item(0x2345, zeros.data)
            ])
        data = packet1 + self.simple_packet()
        item = self.data_to_item(data, 0x1234)
        assert isinstance(item.value, np.uint32)
        assert item.value == 0x12345678

    def test_partial_header(self):
        """Connection closed partway through a header"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:6])
        assert heaps == []

    def test_partial_packet(self):
        """Connection closed partway through item descriptors"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:45])
        assert heaps == []

    def test_partial_payload(self):
        """Connection closed partway through item payload"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:-1])
        assert heaps == []



# Copyright 2015, 2017, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generate and send SPEAD packets.

This is mainly a benchmark application, but also demonstrates the API.
"""

import logging
import sys
import time
import itertools
import argparse
import collections
import asyncio

import numpy as np

import spead2
import spead2.send
import spead2.send.asyncio
from . import cmdline


def parse_endpoint(endpoint):
    if ':' not in endpoint:
        raise ValueError('destination must have the form <host>:<port>')
    return cmdline.parse_endpoint(endpoint)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('destination', type=parse_endpoint, nargs='+', metavar='HOST:PORT')

    group = parser.add_argument_group('Data options')
    group.add_argument('--heap-size', metavar='BYTES', type=int, default=4194304,
                       help='Payload size for heap [%(default)s]')
    group.add_argument('--items', type=int, default=1,
                       help='Number of items per heap [%(default)s]')
    group.add_argument('--dtype', type=str, default='<c8',
                       help='Numpy data type [%(default)s]')
    group.add_argument('--heaps', type=int,
                       help='Number of data heaps to send [infinite]')

    group = parser.add_argument_group('Output options')
    group.add_argument('--log', metavar='LEVEL', default='INFO',
                       help='Log level [%(default)s]')

    group = parser.add_argument_group('Protocol and performance options')
    protocol = cmdline.ProtocolOptions()
    sender = cmdline.SenderOptions(protocol)
    protocol.add_arguments(group)
    sender.add_arguments(group)
    group.add_argument('--descriptors', type=int,
                       help='Description issue frequency [only at start]')

    args = parser.parse_args()
    protocol.notify(parser, args)
    sender.notify(parser, args)
    if protocol.tcp and len(args.destination) > 1:
        parser.error('only one destination is supported with TCP')
    return args, sender


async def run(item_group, stream, args):
    n_bytes = 0
    n_errors = 0
    last_error = None
    start_time = time.time()
    tasks = collections.deque()
    if args.heaps is None:
        rep = itertools.repeat(False)
    else:
        rep = itertools.chain(
            itertools.repeat(False, args.heaps),
            itertools.repeat(True, len(args.destination))
        )
    n_substreams = stream.num_substreams
    for i, is_end in enumerate(rep):
        if len(tasks) >= args.max_heaps:
            try:
                n_bytes += await tasks.popleft()
            except Exception as error:
                n_errors += 1
                last_error = error
        if is_end:
            heap = item_group.get_end()
        else:
            heap = item_group.get_heap(
                descriptors='all' if i < n_substreams else 'stale',
                data='all')
        task = asyncio.ensure_future(stream.async_send_heap(heap, substream_index=i % n_substreams))
        tasks.append(task)
    while len(tasks) > 0:
        try:
            n_bytes += await tasks.popleft()
        except Exception as error:
            n_errors += 1
            last_error = error
    elapsed = time.time() - start_time
    if last_error is not None:
        logging.warn('%d errors, last one: %s', n_errors, last_error)
    print('Sent {} bytes in {:.6f}s, {:.6f} Gb/s'.format(
        n_bytes, elapsed, n_bytes * 8 / elapsed / 1e9))


async def async_main():
    args, sender = get_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    dtype = np.dtype(args.dtype)
    elements = args.heap_size // (args.items * dtype.itemsize)
    heap_size = elements * args.items * dtype.itemsize
    if heap_size != args.heap_size:
        logging.warn('Heap size is not an exact multiple: using %d instead of %d',
                     heap_size, args.heap_size)
    item_group = spead2.send.ItemGroup(
        descriptor_frequency=args.descriptors,
        flavour=sender.make_flavour())
    for i in range(args.items):
        item_group.add_item(id=None, name=f'Test item {i}',
                            description='A test item with arbitrary value',
                            shape=(elements,), dtype=dtype,
                            value=np.zeros((elements,), dtype=dtype))
    thread_pool = sender.make_thread_pool()
    memory_regions = [item.value for item in item_group.values()]
    stream = await sender.make_stream(thread_pool, args.destination, memory_regions)

    await run(item_group, stream, args)


def main():
    try:
        asyncio.get_event_loop().run_until_complete(async_main())
    except KeyboardInterrupt:
        sys.exit(1)



#!/usr/bin/env python

# Copyright 2015, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import asyncio

import numpy as np

import spead2
import spead2.send
import spead2.send.asyncio


logging.basicConfig(level=logging.INFO)

thread_pool = spead2.ThreadPool()
stream = spead2.send.asyncio.UdpStream(
    thread_pool, [("127.0.0.1", 8888)], spead2.send.StreamConfig(rate=1e7))
del thread_pool  # Make sure this doesn't crash anything

shape = (40, 50)
ig = spead2.send.ItemGroup(flavour=spead2.Flavour(4, 64, 48, 0))
item = ig.add_item(0x1234, 'foo', 'a foo item', shape=shape, dtype=np.int32)
item.value = np.zeros(shape, np.int32)
futures = [
    stream.async_send_heap(ig.get_heap()),
    stream.async_send_heap(ig.get_end())
]
# Delete things to check that there are no refcounting bugs
del ig
del stream
asyncio.get_event_loop().run_until_complete(asyncio.wait(futures))



# Copyright 2015, 2017, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Benchmark tool to estimate the sustainable SPEAD bandwidth between two
machines, for a specific set of configurations.

Since UDP is lossy, this is not a trivial problem. We binary search for the
speed that is just sustainable. To make the test at a specific speed more
reliable, it is repeated several times, opening a new stream each time, and
with a delay to allow processors to return to idle states. A TCP control
stream is used to synchronise the two ends. All configuration is done on
the master end.
"""

import sys
import argparse
import json
import collections
import logging
import traceback
import timeit
import asyncio

import numpy as np

import spead2
import spead2.recv
import spead2.recv.asyncio
import spead2.send
import spead2.send.asyncio
from . import cmdline


class AgentConnection:
    def __init__(self, reader, writer):
        self.reader = reader
        self.writer = writer

    async def run_stream(self, stream):
        num_heaps = 0
        while True:
            try:
                await stream.get()
                num_heaps += 1
            except spead2.Stopped:
                return num_heaps

    def _write(self, s):
        self.writer.write(s.encode('ascii'))

    async def run_control(self):
        try:
            stream_task = None
            while True:
                command = await self.reader.readline()
                command = command.decode('ascii')
                logging.debug("command = %s", command)
                if not command:
                    break
                command = json.loads(command)
                if command['cmd'] == 'start':
                    if stream_task is not None:
                        logging.warning("Start received while already running: %s", command)
                        continue
                    args = argparse.Namespace(**command['args'])
                    protocol = cmdline.ProtocolOptions()
                    receiver = cmdline.ReceiverOptions(protocol)
                    for (key, value) in command['protocol'].items():
                        setattr(protocol, key, value)
                    for (key, value) in command['receiver'].items():
                        setattr(receiver, key, value)
                    stream = spead2.recv.asyncio.Stream(
                        receiver.make_thread_pool(),
                        receiver.make_stream_config(),
                        receiver.make_ring_stream_config()
                    )
                    if getattr(receiver, 'ibv') and not hasattr(stream, 'add_udp_ibv_reader'):
                        logging.error('--recv-ibv passed but agent does not support ibv')
                        sys.exit(1)

                    endpoint = args.multicast or args.endpoint
                    receiver.add_readers(stream, [endpoint])
                    stream_task = asyncio.ensure_future(self.run_stream(stream))
                    self._write('ready\n')
                elif command['cmd'] == 'stop':
                    if stream_task is None:
                        logging.warning("Stop received when already stopped")
                        continue
                    stream.stop()
                    received_heaps = await stream_task
                    self._write(json.dumps({'received_heaps': received_heaps}) + '\n')
                    stream_task = None
                    stream = None
                elif command['cmd'] == 'exit':
                    break
                else:
                    logging.warning("Bad command: %s", command)
                    continue
            logging.debug("Connection closed")
            if stream_task is not None:
                stream.stop()
                await stream_task
        except Exception:
            traceback.print_exc()


async def agent_connection(reader, writer):
    try:
        conn = AgentConnection(reader, writer)
        await conn.run_control()
    except Exception:
        traceback.print_exc()


async def run_agent(args):
    server = await asyncio.start_server(agent_connection, port=args.port)
    await server.wait_closed()


async def send_stream(item_group, stream, num_heaps, args):
    tasks = collections.deque()
    transferred = 0
    for i in range(num_heaps + 1):
        while len(tasks) >= args.max_heaps:
            transferred += await tasks.popleft()
        if i == num_heaps:
            heap = item_group.get_end()
        else:
            heap = item_group.get_heap(data='all')
        task = asyncio.ensure_future(stream.async_send_heap(heap))
        tasks.append(task)
    for task in tasks:
        transferred += await task
    return transferred


async def measure_connection_once(args, protocol, sender, receiver,
                                  rate, num_heaps, required_heaps):
    def write(s):
        writer.write(s.encode('ascii'))

    host, port = cmdline.parse_endpoint(args.endpoint)
    reader, writer = await asyncio.open_connection(host, port)
    write(json.dumps(
        {
            'cmd': 'start',
            'args': vars(args),
            'protocol': {key: value for (key, value) in protocol.__dict__.items()
                         if not key.startswith('_')},
            'receiver': {key: value for (key, value) in receiver.__dict__.items()
                         if not key.startswith('_')}
        }) + '\n')
    # Wait for "ready" response
    response = await reader.readline()
    assert response == b'ready\n'

    item_group = spead2.send.ItemGroup(flavour=sender.make_flavour())
    item_group.add_item(id=None, name='Test item',
                        description='A test item with arbitrary value',
                        shape=(args.heap_size,), dtype=np.uint8,
                        value=np.zeros((args.heap_size,), dtype=np.uint8))

    sender.rate = rate * 8e-9        # Convert to Gb/s
    endpoint = args.endpoint
    if args.multicast is not None:
        endpoint = args.multicast
    memory_regions = [item.value for item in item_group.values()]
    stream = await sender.make_stream(
        sender.make_thread_pool(), [cmdline.parse_endpoint(endpoint)], memory_regions)

    start = timeit.default_timer()
    transferred = await send_stream(item_group, stream, num_heaps, args)
    end = timeit.default_timer()
    elapsed = end - start
    actual_rate = transferred / elapsed
    # Give receiver time to catch up with any queue
    await asyncio.sleep(0.1)
    write(json.dumps({'cmd': 'stop'}) + '\n')
    # Read number of heaps received
    response = await reader.readline()
    response = json.loads(response.decode('ascii'))
    received_heaps = response['received_heaps']
    await asyncio.sleep(0.5)
    await writer.drain()
    writer.close()
    logging.debug("Received %d/%d heaps in %f seconds, rate %.3f Gbps",
                  received_heaps, num_heaps, elapsed, actual_rate * 8e-9)
    return received_heaps >= required_heaps, actual_rate


async def measure_connection(args, protocol, sender, receiver, rate, num_heaps, required_heaps):
    good = True
    rate_sum = 0.0
    passes = 5
    for i in range(passes):
        status, actual_rate = await measure_connection_once(args, protocol, sender, receiver,
                                                            rate, num_heaps, required_heaps)
        good = good and status
        rate_sum += actual_rate
    return good, rate_sum / passes


async def run_master(args, protocol, sender, receiver):
    best_actual = 0.0

    # Send 1GB as fast as possible to find an upper bound - receive rate
    # does not matter. Also do a warmup run first to warm up the receiver.
    num_heaps = int(1e9 / args.heap_size) + 2
    await measure_connection_once(args, protocol, sender, receiver, 0.0, num_heaps, 0)  # warmup
    good, actual_rate = await measure_connection(args, protocol, sender, receiver,
                                                 0.0, num_heaps, num_heaps - 1)
    if good:
        if not args.quiet:
            print("Limited by send spead")
        best_actual = actual_rate
    else:
        print("Send rate: {:.3f} Gbps".format(actual_rate * 8e-9))
        low = 0.0
        high = actual_rate
        while high - low > high * 0.02:
            # Need at least 1GB of data to overwhelm cache effects, and want at least
            # 1 second for warmup effects.
            rate = (low + high) * 0.5
            num_heaps = int(max(1e9, rate) / args.heap_size) + 2
            good, actual_rate = await measure_connection(
                args, protocol, sender, receiver, rate, num_heaps, num_heaps - 1)
            if not args.quiet:
                print("Rate: {:.3f} Gbps ({:.3f} actual): {}".format(
                    rate * 8e-9, actual_rate * 8e-9, "GOOD" if good else "BAD"))
            if good:
                low = rate
                best_actual = actual_rate
            else:
                high = rate
    rate_gbps = best_actual * 8e-9
    if args.quiet:
        print(rate_gbps)
    else:
        print(f"Sustainable rate: {rate_gbps:.3f} Gbps")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--log', metavar='LEVEL', default='INFO', help='Log level [%(default)s]')
    subparsers = parser.add_subparsers(title='subcommands')

    master = subparsers.add_parser('master')
    sender_map = {
        'buffer': 'send_buffer',
        'packet': None,                 # Use receiver's packet size
        'rate': None,                   # Controlled by test
        'bind': 'send_bind',
        'ibv': 'send_ibv',
        'ibv_vector': 'send_ibv_vector',
        'ibv_max_poll': 'send_ibv_max_poll',
        'affinity': 'send-affinity',
        'threads': 'send-threads'
    }
    receiver_map = {
        'buffer': 'recv_buffer',
        'bind': 'recv_bind',
        'ibv': 'recv_ibv',
        'ibv_vector': 'recv_ibv_vector',
        'ibv_max_poll': 'recv_ibv_max_poll',
        'affinity': 'recv-affinity',
        'threads': 'recv-threads',
        'mem_pool': None,
        'mem_lower': None,
        'mem_upper': None
    }
    protocol = cmdline.ProtocolOptions(name_map={'tcp': None})
    sender = cmdline.SenderOptions(protocol, name_map=sender_map)
    receiver = cmdline.ReceiverOptions(protocol, name_map=receiver_map)
    master.add_argument('--quiet', action='store_true', default=False,
                        help='Print only the final result')
    master.add_argument('--heap-size', metavar='BYTES', type=int, default=4194304,
                        help='Payload size for heap [%(default)s]')
    master.add_argument('--multicast', metavar='ADDRESS', type=str,
                        help='Send via multicast group [unicast]')
    protocol.add_arguments(master)
    sender.add_arguments(master.add_argument_group('sender options'))
    receiver.add_arguments(master.add_argument_group('receiver options'))
    master.add_argument('endpoint', metavar='HOST:PORT')
    agent = subparsers.add_parser('agent')
    agent.add_argument('port', type=int)

    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))
    if 'endpoint' in args:
        if args.send_ibv and not args.multicast:
            parser.error('--send-ibv requires --multicast')
        receiver.mem_pool = True
        receiver.mem_lower = args.heap_size
        receiver.mem_upper = args.heap_size + 1024  # more than enough for overheads
        protocol.notify(parser, args)
        receiver.notify(parser, args)
        sender.packet = receiver.packet
        sender.notify(parser, args)
        task = run_master(args, protocol, sender, receiver)
    else:
        task = run_agent(args)
    task = asyncio.ensure_future(task)
    asyncio.get_event_loop().run_until_complete(task)
    task.result()



# Copyright 2015, 2019-2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test that data can be passed over the SPEAD protocol using the various transports."""

import os
import socket
import sys

import numpy as np
import netifaces
import pytest

import spead2
import spead2.send
import spead2.recv


def _assert_items_equal(item1, item2):
    assert item1.id == item2.id
    assert item1.name == item2.name
    assert item1.description == item2.description
    assert item1.shape == item2.shape
    assert item1.format == item2.format
    # Byte order need not match, provided that values are received correctly
    if item1.dtype is not None and item2.dtype is not None:
        assert item1.dtype.newbyteorder('<') == item2.dtype.newbyteorder('<')
    else:
        assert item1.dtype == item2.dtype
    assert item1.order == item2.order
    # Comparing arrays has many issues. Convert them to lists where appropriate
    value1 = item1.value
    value2 = item2.value
    if hasattr(value1, 'tolist'):
        value1 = value1.tolist()
    if hasattr(value2, 'tolist'):
        value2 = value2.tolist()
    assert value1 == value2


def assert_item_groups_equal(item_group1, item_group2):
    assert sorted(item_group1.keys()) == sorted(item_group2.keys())
    for key in item_group1.keys():
        _assert_items_equal(item_group1[key], item_group2[key])


@pytest.mark.timeout(5)
class BaseTestPassthrough:
    """Tests common to all transports and libraries"""

    is_lossy = False
    requires_ipv6 = False

    @classmethod
    def check_ipv6(cls):
        if not socket.has_ipv6:
            pytest.skip('platform does not support IPv6')
        # Travis build systems fail to bind to an IPv6 address
        sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        try:
            sock.bind(("::1", 8888))
        except OSError:
            pytest.skip('platform cannot bind IPv6 localhost address')
        finally:
            sock.close()

    def _test_item_groups(self, item_groups,
                          memcpy=spead2.MEMCPY_STD, allocator=None, new_order='='):
        received_item_groups = self.transmit_item_groups(item_groups, memcpy, allocator, new_order)
        assert len(received_item_groups) == len(item_groups)
        for received_item_group, item_group in zip(received_item_groups, item_groups):
            assert_item_groups_equal(item_group, received_item_group)
            for item in received_item_group.values():
                if item.dtype is not None:
                    assert item.value.dtype == item.value.dtype.newbyteorder(new_order)

    def _test_item_group(self, item_group,
                         memcpy=spead2.MEMCPY_STD, allocator=None, new_order='='):
        self._test_item_groups([item_group])

    def test_numpy_simple(self):
        """A basic array with numpy encoding"""
        ig = spead2.send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        self._test_item_group(ig)

    def test_numpy_byteorder(self):
        """A numpy array in non-native byte order"""
        ig = spead2.send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.dtype(np.uint16).newbyteorder())
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        self._test_item_group(ig)
        self._test_item_group(ig, new_order='|')

    def test_numpy_large(self):
        """A numpy style array split across several packets. It also
        uses non-temporal copies, a custom allocator, and a memory pool,
        to test that those all work."""
        # macOS doesn't have a big enough socket buffer to reliably transmit
        # the whole thing over UDP.
        if self.is_lossy and sys.platform == 'darwin':
            pytest.skip("macOS can't reliably handle large heaps over UDP")
        ig = spead2.send.ItemGroup()
        data = np.random.randn(100, 200)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        allocator = spead2.MmapAllocator()
        pool = spead2.MemoryPool(1, 4096, 4, 4, allocator)
        self._test_item_group(ig, spead2.MEMCPY_NONTEMPORAL, pool)

    def test_fallback_struct_whole_bytes(self):
        """A structure with non-byte-aligned elements, but which is
        byte-aligned overall."""
        ig = spead2.send.ItemGroup()
        format = [('u', 4), ('f', 64), ('i', 4)]
        data = (12, 1.5, -3)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(), format=format, value=data)
        self._test_item_group(ig)

    def test_string(self):
        """Byte string is converted to array of characters and back."""
        ig = spead2.send.ItemGroup()
        format = [('c', 8)]
        data = 'Hello world'
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(None,), format=format, value=data)
        self._test_item_group(ig)

    def test_fallback_array_partial_bytes_small(self):
        """An array which takes a fractional number of bytes per element
        and is small enough to encode in an immediate.
        """
        ig = spead2.send.ItemGroup()
        format = [('u', 7)]
        data = [127, 12, 123]
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(len(data),), format=format, value=data)
        self._test_item_group(ig)

    def test_fallback_types(self):
        """An array structure using a mix of types."""
        ig = spead2.send.ItemGroup()
        format = [('b', 1), ('i', 7), ('c', 8), ('f', 32)]
        data = [(True, 17, b'y', 1.0), (False, -23, b'n', -1.0)]
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(2,), format=format, value=data)
        self._test_item_group(ig)

    def test_numpy_fallback_struct(self):
        """A structure specified using a format, but which is encodable using numpy."""
        ig = spead2.send.ItemGroup()
        format = [('u', 8), ('f', 32)]
        data = (12, 1.5)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(), format=format, value=data)
        self._test_item_group(ig)

    def test_fallback_struct_partial_bytes(self):
        """A structure which takes a fractional number of bytes per element."""
        ig = spead2.send.ItemGroup()
        format = [('u', 4), ('f', 64)]
        data = (12, 1.5)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=(), format=format, value=data)
        self._test_item_group(ig)

    def test_fallback_scalar(self):
        """Send a scalar using fallback format descriptor."""
        ig = spead2.send.ItemGroup()
        format = [('f', 64)]
        data = 1.5
        ig.add_item(id=0x2345, name='scalar name', description='scalar description',
                    shape=(), format=format, value=data)
        self._test_item_group(ig)

    def test_many_items(self):
        """Sends many items.

        The implementation handles few-item heaps differently (for
        performance), so need to test that the many-item code works too.
        """
        ig = spead2.send.ItemGroup()
        for i in range(50):
            name = f'test item {i}'
            ig.add_item(id=0x2345 + i, name=name, description=name,
                        shape=(), format=[('u', 40)], value=0x12345 * i)
        self._test_item_group(ig)

    def transmit_item_groups(self, item_groups, memcpy, allocator, new_order='='):
        """Transmit `item_groups` over the chosen transport.

        Return the item groups received at the other end. Each item group will
        be transmitted over a separate substream (thus, the transport must
        support substreams if `item_groups` has more than one element).
        """
        if self.requires_ipv6:
            self.check_ipv6()
        recv_config = spead2.recv.StreamConfig(memcpy=memcpy)
        if allocator is not None:
            recv_config.memory_allocator = allocator
        receivers = [
            spead2.recv.Stream(spead2.ThreadPool(), recv_config)
            for i in range(len(item_groups))
        ]
        self.prepare_receivers(receivers)
        sender = self.prepare_senders(spead2.ThreadPool(), len(item_groups))
        gens = [
            spead2.send.HeapGenerator(item_group)
            for item_group in item_groups
        ]
        if len(item_groups) != 1:
            # Use reversed order so that if everything is actually going
            # through the same transport it will get picked up.
            for i, gen in reversed(list(enumerate(gens))):
                sender.send_heap(gen.get_heap(), substream_index=i)
            for i, gen in enumerate(gens):
                sender.send_heap(gen.get_end(), substream_index=i)
        else:
            # This is a separate code path to give coverage of the case where
            # the substream index is implicit.
            sender.send_heap(gens[0].get_heap())
            sender.send_heap(gens[0].get_end())
        received_item_groups = []
        for receiver in receivers:
            ig = spead2.ItemGroup()
            for heap in receiver:
                ig.update(heap, new_order)
            received_item_groups.append(ig)
        return received_item_groups

    def prepare_receiver(self, receiver):
        """Generate a single receiver to use in the test.

        This should be implemented by classes that don't support substreams.
        """
        raise NotImplementedError()

    def prepare_sender(self, thread_pool):
        """Generate a sender with a single substream.

        This should be implemented by classes that don't support substreams.
        """
        raise NotImplementedError()

    def prepare_receivers(self, receivers):
        """Generate receivers to use in the test."""
        assert len(receivers) == 1
        self.prepare_receiver(receivers[0])

    def prepare_senders(self, thread_pool, n):
        """Generate a sender to use in the test, with `n` substreams."""
        assert n == 1
        return self.prepare_sender(thread_pool)


class BaseTestPassthroughSubstreams(BaseTestPassthrough):
    """Tests for send stream classes that support multiple substreams."""

    def test_substreams(self):
        item_groups = []
        for i in range(4):
            ig = spead2.ItemGroup()
            ig.add_item(id=0x2345, name='int', description='an integer',
                        shape=(), format=[('i', 32)], value=i)
            item_groups.append(ig)
        self._test_item_groups(item_groups)

    def prepare_receivers(self, receivers):
        raise NotImplementedError()

    def prepare_senders(self, thread_pool, n):
        raise NotImplementedError()


class TestPassthroughUdp(BaseTestPassthroughSubstreams):
    is_lossy = True

    def prepare_receivers(self, receivers):
        for i, receiver in enumerate(receivers):
            receiver.add_udp_reader(8888 + i, bind_hostname="localhost")

    def prepare_senders(self, thread_pool, n):
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.UdpStream(
                    thread_pool, "localhost", 8888,
                    spead2.send.StreamConfig(rate=1e7),
                    buffer_size=0)
        else:
            return spead2.send.UdpStream(
                thread_pool,
                [("localhost", 8888 + i) for i in range(n)],
                spead2.send.StreamConfig(rate=1e7),
                buffer_size=0)

    def test_empty_endpoints(self):
        with pytest.raises(ValueError):
            spead2.send.UdpStream(
                spead2.ThreadPool(), [], spead2.send.StreamConfig(rate=1e7))

    def test_mixed_protocols(self):
        with pytest.raises(ValueError):
            spead2.send.UdpStream(
                spead2.ThreadPool(),
                [('127.0.0.1', 8888), ('::1', 8888)],
                spead2.send.StreamConfig(rate=1e7))


class TestPassthroughUdp6(BaseTestPassthroughSubstreams):
    is_lossy = True
    requires_ipv6 = True

    def prepare_receivers(self, receivers):
        for i, receiver in enumerate(receivers):
            receiver.add_udp_reader(8888 + i, bind_hostname="::1")

    def prepare_senders(self, thread_pool, n):
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.UdpStream(
                    thread_pool, "::1", 8888,
                    spead2.send.StreamConfig(rate=1e7),
                    buffer_size=0)
        else:
            return spead2.send.UdpStream(
                thread_pool,
                [("::1", 8888 + i) for i in range(n)],
                spead2.send.StreamConfig(rate=1e7),
                buffer_size=0)


class TestPassthroughUdpCustomSocket(BaseTestPassthroughSubstreams):
    is_lossy = True

    def prepare_receivers(self, receivers):
        self._ports = []
        for i, receiver in enumerate(receivers):
            recv_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            recv_sock.bind(("127.0.0.1", 0))
            self._ports.append(recv_sock.getsockname()[1])
            receiver.add_udp_reader(socket=recv_sock)
            recv_sock.close()   # spead2 duplicates the socket

    def prepare_senders(self, thread_pool, n):
        assert len(self._ports) == n
        send_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if n == 1:
            with pytest.deprecated_call():
                sender = spead2.send.UdpStream(
                    thread_pool, send_sock, "127.0.0.1", self._ports[0],
                    spead2.send.StreamConfig(rate=1e7))
        else:
            sender = spead2.send.UdpStream(
                thread_pool, send_sock,
                [("127.0.0.1", port) for port in self._ports],
                spead2.send.StreamConfig(rate=1e7))
        send_sock.close()   # spead2 duplicates the socket
        return sender


class TestPassthroughUdpMulticast(BaseTestPassthroughSubstreams):
    is_lossy = True
    MCAST_GROUP = '239.255.88.88'
    INTERFACE_ADDRESS = '127.0.0.1'

    def prepare_receivers(self, receivers):
        for i, receiver in enumerate(receivers):
            receiver.add_udp_reader(
                self.MCAST_GROUP, 8887 - i, interface_address=self.INTERFACE_ADDRESS)

    def prepare_senders(self, thread_pool, n):
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.UdpStream(
                    thread_pool, self.MCAST_GROUP, 8887,
                    spead2.send.StreamConfig(rate=1e7),
                    buffer_size=0, ttl=1, interface_address=self.INTERFACE_ADDRESS)
        else:
            return spead2.send.UdpStream(
                thread_pool,
                [(self.MCAST_GROUP, 8887 - i) for i in range(n)],
                spead2.send.StreamConfig(rate=1e7),
                buffer_size=0, ttl=1, interface_address=self.INTERFACE_ADDRESS)


class TestPassthroughUdp6Multicast(TestPassthroughUdp6):
    MCAST_GROUP = 'ff14::1234'

    @classmethod
    def get_interface_index(cls):
        for iface in netifaces.interfaces():
            addrs = netifaces.ifaddresses(iface).get(netifaces.AF_INET6, [])
            for addr in addrs:
                if addr['addr'] != '::1':
                    return socket.if_nametoindex(iface)
        pytest.skip('could not find suitable interface for test')

    def prepare_receivers(self, receivers):
        interface_index = self.get_interface_index()
        for i, receiver in enumerate(receivers):
            receiver.add_udp_reader(self.MCAST_GROUP, 8887 - i, interface_index=interface_index)

    def prepare_senders(self, thread_pool, n):
        interface_index = self.get_interface_index()
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.UdpStream(
                    thread_pool, self.MCAST_GROUP, 8887,
                    spead2.send.StreamConfig(rate=1e7),
                    buffer_size=0, ttl=0, interface_index=interface_index)
        else:
            return spead2.send.UdpStream(
                thread_pool,
                [(self.MCAST_GROUP, 8887 - i) for i in range(n)],
                spead2.send.StreamConfig(rate=1e7),
                buffer_size=0, ttl=0, interface_index=interface_index)


class TestPassthroughUdpIbv(BaseTestPassthroughSubstreams):
    is_lossy = True
    MCAST_GROUP = '239.255.88.88'

    def _interface_address(self):
        ifaddr = os.getenv('SPEAD2_TEST_IBV_INTERFACE_ADDRESS')
        if not ifaddr:
            pytest.skip('Envar SPEAD2_TEST_IBV_INTERFACE_ADDRESS not set')
        return ifaddr

    def setup(self):
        # mlx5 drivers only enable multicast loopback if there are multiple
        # device contexts. The sender and receiver end up sharing one, so we
        # need to explicitly create another.
        if not hasattr(spead2, 'IbvContext'):
            pytest.skip('IBV support not compiled in')
        self._extra_context = spead2.IbvContext(self._interface_address())

    def teardown(self):
        self._extra_context.reset()

    def prepare_receivers(self, receivers):
        for i, receiver in enumerate(receivers):
            receiver.add_udp_ibv_reader(
                spead2.recv.UdpIbvConfig(
                    endpoints=[(self.MCAST_GROUP, 8876 + i)],
                    interface_address=self._interface_address()))

    def prepare_senders(self, thread_pool, n):
        # The buffer size is deliberately reduced so that we test the
        # wrapping once the buffer has been used up
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.UdpIbvStream(
                    thread_pool, self.MCAST_GROUP, 8876,
                    spead2.send.StreamConfig(rate=1e7),
                    self._interface_address(),
                    buffer_size=64 * 1024)
        else:
            return spead2.send.UdpIbvStream(
                thread_pool,
                spead2.send.StreamConfig(rate=1e7),
                spead2.send.UdpIbvConfig(
                    endpoints=[(self.MCAST_GROUP, 8876 + i) for i in range(n)],
                    interface_address=self._interface_address(),
                    buffer_size=64 * 1024
                )
            )

    @pytest.mark.parametrize('num_items', [0, 1, 3, 4, 10])
    def test_memory_regions(self, num_items):
        receiver = spead2.recv.Stream(spead2.ThreadPool(), spead2.recv.StreamConfig())
        receiver.add_udp_ibv_reader(
            spead2.recv.UdpIbvConfig(
                endpoints=[(self.MCAST_GROUP, 8876)],
                interface_address=self._interface_address()))

        ig = spead2.send.ItemGroup()
        data = [np.random.randn(50) for i in range(num_items)]
        for i in range(num_items):
            ig.add_item(id=0x2345 + i, name=f'name {i}', description=f'description {i}',
                        shape=data[i].shape, dtype=data[i].dtype, value=data[i])
        sender = spead2.send.UdpIbvStream(
            spead2.ThreadPool(),
            spead2.send.StreamConfig(rate=1e7),
            spead2.send.UdpIbvConfig(
                endpoints=[(self.MCAST_GROUP, 8876)],
                interface_address=self._interface_address(),
                memory_regions=data
            )
        )
        sender.send_heap(ig.get_heap())
        sender.send_heap(ig.get_end())

        recv_ig = spead2.ItemGroup()
        for heap in receiver:
            recv_ig.update(heap)
        assert_item_groups_equal(ig, recv_ig)


class TestPassthroughTcp(BaseTestPassthrough):
    def prepare_receiver(self, receiver):
        receiver.add_tcp_reader(8887, bind_hostname="127.0.0.1")

    def prepare_sender(self, thread_pool):
        return spead2.send.TcpStream(thread_pool, [("127.0.0.1", 8887)])


class TestPassthroughTcpCustomSocket(BaseTestPassthrough):
    def prepare_receiver(self, receiver):
        sock = socket.socket()
        sock.bind(("127.0.0.1", 0))
        self._port = sock.getsockname()[1]
        sock.listen(1)
        receiver.add_tcp_reader(sock)

    def prepare_sender(self, thread_pool):
        sock = socket.socket()
        sock.connect(("127.0.0.1", self._port))
        return spead2.send.TcpStream(thread_pool, sock)


class TestPassthroughTcp6(BaseTestPassthrough):
    requires_ipv6 = True

    def prepare_receiver(self, receiver):
        receiver.add_tcp_reader(8887, bind_hostname="::1")

    def prepare_sender(self, thread_pool):
        with pytest.deprecated_call():
            return spead2.send.TcpStream(thread_pool, "::1", 8887)


class TestPassthroughMem(BaseTestPassthrough):
    def transmit_item_groups(self, item_groups, memcpy, allocator, new_order='='):
        assert len(item_groups) == 1
        thread_pool = spead2.ThreadPool(2)
        sender = spead2.send.BytesStream(thread_pool)
        gen = spead2.send.HeapGenerator(item_groups[0])
        sender.send_heap(gen.get_heap())
        sender.send_heap(gen.get_end())
        recv_config = spead2.recv.StreamConfig(memcpy=memcpy)
        if allocator is not None:
            recv_config.memory_allocator = allocator
        receiver = spead2.recv.Stream(thread_pool, recv_config)
        receiver.add_buffer_reader(sender.getvalue())
        received_item_group = spead2.ItemGroup()
        for heap in receiver:
            received_item_group.update(heap, new_order)
        return [received_item_group]


class TestPassthroughInproc(BaseTestPassthroughSubstreams):
    def prepare_receivers(self, receivers):
        assert len(receivers) == len(self._queues)
        for receiver, queue in zip(receivers, self._queues):
            receiver.add_inproc_reader(queue)

    def prepare_senders(self, thread_pool, n):
        assert n == len(self._queues)
        if n == 1:
            with pytest.deprecated_call():
                return spead2.send.InprocStream(thread_pool, self._queues[0])
        else:
            return spead2.send.InprocStream(thread_pool, self._queues)

    def transmit_item_groups(self, item_groups, memcpy, allocator, new_order='='):
        self._queues = [spead2.InprocQueue() for ig in item_groups]
        ret = super().transmit_item_groups(
            item_groups, memcpy, allocator, new_order)
        for queue in self._queues:
            queue.stop()
        return ret

    def test_queues(self):
        queues = [spead2.InprocQueue() for i in range(2)]
        stream = spead2.send.InprocStream(spead2.ThreadPool(), queues)
        assert stream.queues == queues
        with pytest.deprecated_call():
            with pytest.raises(RuntimeError):
                stream.queue

    def test_queue(self):
        queue = spead2.InprocQueue()
        stream = spead2.send.InprocStream(spead2.ThreadPool(), [queue])
        assert stream.queues == [queue]
        with pytest.deprecated_call():
            assert stream.queue is queue



#!/usr/bin/env python3

# Copyright 2020 National Research Foundation (SARAO)
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import re
import sys
from distutils.version import StrictVersion


parser = argparse.ArgumentParser()
parser.add_argument('mode', choices=('major', 'minor', 'patch', 'full'))
args = parser.parse_args()

with open('src/spead2/_version.py') as version_file:
    line = version_file.readline().strip()
match = re.fullmatch(r'__version__ = "(.+)"', line)
if not match:
    print('src/spead2/_version.py does not match the expected format', file=sys.stderr)
    sys.exit(1)
version_str = match.group(1)
version = StrictVersion(version_str)

mode = args.mode
if mode == 'major':
    print(version.version[0], end='')
elif mode == 'minor':
    print(version.version[1], end='')
elif mode == 'patch':
    print(version.version[2], end='')
else:
    print(version_str, end='')
